#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint sign mappings** (i.e., dictionary globals mapping from
instances of the :class:`beartype._data.hint.pep.sign.datapepsigncls.HintSign`
class to various metadata associated with categories of type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Dict
from beartype._data.hint.pep.sign.datapepsigncls import HintSign
from beartype._data.hint.pep.sign.datapepsigns import (
    HintSignAbstractSet,
    HintSignAsyncContextManager,
    HintSignAsyncGenerator,
    HintSignAsyncIterator,
    HintSignAsyncIterable,
    HintSignAwaitable,
    # HintSignByteString,
    HintSignChainMap,
    HintSignCollection,
    HintSignContainer,
    HintSignContextManager,
    HintSignCoroutine,
    HintSignCounter,
    HintSignDefaultDict,
    HintSignDeque,
    HintSignDict,
    HintSignFrozenSet,
    HintSignGenerator,
    HintSignItemsView,
    HintSignIterable,
    HintSignIterator,
    HintSignKeysView,
    HintSignList,
    HintSignMapping,
    HintSignMappingView,
    HintSignMatch,
    HintSignMutableMapping,
    HintSignMutableSequence,
    HintSignMutableSet,
    HintSignOrderedDict,
    HintSignPattern,
    HintSignReversible,
    HintSignSequence,
    HintSignSet,
    HintSignTuple,
    HintSignType,
    HintSignValuesView,
)

# ....................{ PRIVATE ~ globals                  }....................
# Note that the builtin "range" class:
# * Initializer "range.__init__(start, stop)" is effectively instantiated as
#   [start, stop) -- that is to say, such that:
#   * The initial "start" integer is *INCLUSIVE* (i.e., the instantiated range
#     includes this integer).
#   * The final "stop" integer is *EXCLUSIVE* (i.e., the instantiated range
#     excludes this integer).
# * Publicizes these integers as the instance variables "start" and "stop" such
#   that these invariants are guaranteed:
#       >>> range(min, max).start == min
#       True
#       >>> range(min, max).stop == max
#       True

_ARGS_LEN_0 = range(0, 1)  # == [0, 1) == [0, 0]
'''
**Zero-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``0``, describing type hint factories subscriptable by
*no* child type hints).
'''


_ARGS_LEN_1 = range(1, 2)  # == [1, 2) == [1, 1]
'''
**One-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``1``, describing type hint factories subscriptable by
exactly one child type hint).
'''


_ARGS_LEN_2 = range(2, 3)  # == [2, 3) == [2, 2]
'''
**Two-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``2``, describing type hint factories subscriptable by
exactly two child type hints).
'''


_ARGS_LEN_3 = range(3, 4)  # == [3, 4) == [3, 3]
'''
**Three-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``3``, describing type hint factories subscriptable by
exactly three child type hints).
'''


_ARGS_LEN_1_OR_2 = range(1, 3)  # == [1, 3) == [1, 2]
'''
**One- or two-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer range ``[1, 2]``, describing type hint factories
subscriptable by either one or two child type hints).
'''

# ....................{ SIGNS ~ origin : args              }....................
# Fully initialized by the _init() function below.
HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE: Dict[HintSign, range] = {
    # Type hint factories subscriptable by exactly one child type hint.
    HintSignAbstractSet: _ARGS_LEN_1,
    HintSignAsyncIterable: _ARGS_LEN_1,
    HintSignAsyncIterator: _ARGS_LEN_1,
    HintSignAwaitable: _ARGS_LEN_1,
    # HintSignByteString: _ARGS_LEN_1,
    HintSignCollection: _ARGS_LEN_1,
    HintSignContainer: _ARGS_LEN_1,
    HintSignCounter: _ARGS_LEN_1,
    HintSignDeque: _ARGS_LEN_1,
    HintSignFrozenSet: _ARGS_LEN_1,
    HintSignIterable: _ARGS_LEN_1,
    HintSignIterator: _ARGS_LEN_1,
    HintSignKeysView: _ARGS_LEN_1,
    HintSignList: _ARGS_LEN_1,
    HintSignMatch: _ARGS_LEN_1,
    HintSignMappingView: _ARGS_LEN_1,
    HintSignMutableSequence: _ARGS_LEN_1,
    HintSignMutableSet: _ARGS_LEN_1,
    HintSignPattern: _ARGS_LEN_1,
    HintSignReversible: _ARGS_LEN_1,
    HintSignSequence: _ARGS_LEN_1,
    HintSignSet: _ARGS_LEN_1,
    HintSignType: _ARGS_LEN_1,
    HintSignValuesView: _ARGS_LEN_1,

    # Type hint factories subscriptable by exactly two child type hints.
    HintSignAsyncGenerator: _ARGS_LEN_2,
    HintSignChainMap: _ARGS_LEN_2,
    HintSignDefaultDict: _ARGS_LEN_2,
    HintSignDict: _ARGS_LEN_2,
    HintSignItemsView: _ARGS_LEN_2,
    HintSignMapping: _ARGS_LEN_2,
    HintSignMutableMapping: _ARGS_LEN_2,
    HintSignOrderedDict: _ARGS_LEN_2,
    HintSignTuple: _ARGS_LEN_2,

    # Type hint factories subscriptable by exactly three child type hints.
    HintSignCoroutine: _ARGS_LEN_3,
    HintSignGenerator: _ARGS_LEN_3,
}
'''
Dictionary mapping from each sign uniquely identifying a PEP-compliant type hint
factory originating from an **isinstanceable origin type** (i.e., isinstanceable
class such that *all* objects satisfying type hints created by subscripting this
factory are instances of this class) to this factory's **argument length range**
(i.e., :class:`range` instance describing the minimum and maximum number of
child type hints that may subscript this factory).
'''

# ....................{ PRIVATE ~ main                     }....................
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    # Defer function-specific imports.
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_13

    # If the active Python interpreter targets Python >= 3.13...
    if IS_PYTHON_AT_LEAST_3_13:
        # Add all signs uniquely identifying one- or two-argument type hint
        # factories under Python >= 3.13, which generalized various one-argument
        # type hint factories to accept an additional optional child type hint
        # via "PEP 696 – Type Defaults for Type Parameters".
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.update({
            HintSignAsyncContextManager: _ARGS_LEN_1_OR_2,
            HintSignContextManager: _ARGS_LEN_1_OR_2,
        })
    # Else, the active Python interpreter targets Python <= 3.12. In this
    # case...
    else:
        # Add all signs uniquely identifying two-argument type hint factories
        # under Python <= 3.12.
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.update({
            HintSignAsyncContextManager: _ARGS_LEN_1,
            HintSignContextManager: _ARGS_LEN_1,
        })
    # print(f'HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE: {HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE}')


# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint reducers** (i.e., low-level callables converting type
hints from one format into another, either losslessly or in a lossy manner).

Type hint reductions imposed by this submodule are purely internal to
:mod:`beartype` itself and thus transient in nature. These reductions are *not*
permanently applied to the ``__annotations__`` dunder dictionaries of the
classes and callables annotated by these type hints.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.meta import URL_ISSUES
from beartype.roar import BeartypeDecorHintReduceException
from beartype.typing import (
    Any,
    Optional,
)
from beartype._check.convert._reduce._redmap import (
    HINT_SIGN_TO_REDUCE_HINT_CACHED_get,
    HINT_SIGN_TO_REDUCE_HINT_UNCACHED_get,
)
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._check.metadata.metasane import (
    HintOrSanifiedData,
    HintSanifiedData,
    unpack_hint_or_sane,
)
from beartype._conf.confcls import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.hint.datahintpep import (
    AnyObject,
    Hint,
    SetHints,
    TypeVarToHint,
)
from beartype._data.hint.datahinttyping import (
    DictStrToAny,
    FrozenSetInts,
    TypeStack,
)
from beartype._data.hint.pep.datapeprepr import HINTS_REPR_IGNORABLE_SHALLOW
from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._data.kind.datakindset import FROZENSET_EMPTY
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.func.arg.utilfuncargiter import ArgKind
from beartype._util.hint.pep.utilpepget import get_hint_pep_sign_or_none
from beartype._util.hint.utilhintget import get_hint_repr
from beartype._util.hint.utilhinttest import die_unless_hint
from beartype._util.kind.map.utilmapfrozen import FrozenDict
from beartype._util.kind.map.utilmapset import remove_mapping_keys
from beartype._util.utilobject import SENTINEL

# ....................{ REDUCERS                           }....................
def reduce_hint_child(
    hint: Hint, kwargs: DictStrToAny) -> HintOrSanifiedData:
    '''
    Lower-level child type hint reduced (i.e., converted) from the passed
    higher-level child type hint if reducible *or* this child type hint as
    is otherwise (i.e., if this child type hint is irreducible).

    This reducer is a convenience wrapper for the more general-purpose
    :func:`.reduce_hint` reducer, simplifying calls to that reducer when passed
    child hints.

    Parameters
    ----------
    hint : Hint
        Child type hint to be reduced.
    kwargs : DictStrToAny
        Keyword parameters to be passed after being unpacked to the lower-level
        :func:`.reduce_hint` reducer. For safety, this reducer silently ignores
        keyword parameters inapplicable to child hints. This includes:

        * ``arg_kind``, applicable *only* to root hints directly annotating
          callable parameters.
        * ``decor_meta``, applicable *only* to root hints directly annotating
          callable parameters or returns.
        * ``pith_name``, applicable *only* to root hints directly annotating
          callable parameters or returns.

    Returns
    -------
    HintOrSanifiedData
        Either:

        * If this hint is reducible, either:

          * If reducing this hint to a lower-level hint generates supplementary
            metadata, that metadata including that lower-level hint.
          * Else, that lower-level hint alone.

        * Else, this hint is irreducible. In this case, this hint unmodified.
    '''

    # Remove all unsafe keyword parameters (i.e., parameters that are
    # inapplicable to child hints and thus *NOT* safely passable to the
    # subsequently called reduce_hint() function) from this dictionary.
    remove_mapping_keys(kwargs, _REDUCE_HINT_CHILD_ARG_NAMES_UNSAFE)

    # Return this child hint possibly reduced to a lower-level hint.
    return reduce_hint(hint=hint, **kwargs)


def reduce_hint(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    arg_kind: Optional[ArgKind] = None,
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    decor_meta: Optional[BeartypeDecorMeta] = None,
    hints_overridden: Optional[SetHints] = None,
    pith_name: Optional[str] = None,
    recursable_hint_ids: FrozenSetInts = FROZENSET_EMPTY,
    reductions_count: int = 0,
    typevar_to_hint: TypeVarToHint = FROZENDICT_EMPTY,
    exception_prefix: str = '',
) -> HintOrSanifiedData:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible *or* this hint as is otherwise (i.e., if
    this hint is irreducible).

    This reducer *cannot* be meaningfully memoized, since multiple passed
    parameters (e.g., ``pith_name``, ``cls_stack``) are typically isolated to a
    handful of callables across the codebase currently being decorated by
    :mod:`beartype`. Memoizing this reducer would needlessly consume space and
    time. To improve efficiency, this reducer is instead implemented in terms of
    two lower-level private reducers:

    * The memoized :func:`._reduce_hint_cached` reducer, responsible for
      efficiently reducing *most* (but not all) type hints.
    * The unmemoized :func:`._reduce_hint_uncached` reducer, responsible for
      inefficiently reducing the small subset of type hints contextually
      requiring these problematic parameters.

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.

        Defaults to :data:`None`.
    cls_stack : TypeStack, optional
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to the default beartype configuration.
    decor_meta : Optional[BeartypeDecorMeta], optional
        Either:

        * If this hint annotates a parameter or return of some callable, the
          :mod:`beartype`-specific decorator metadata describing that callable.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    hints_overridden : set[Hint], optional
        Mutable set of all previously **overridden hints** (i.e., hints reduced
        to user-defined hint overrides configured by the
        :attr:`beartype.BeartypeConf.hint_overrides` dictionary) reduced by a
        prior call to this function, guarding against accidental infinite
        recursion between lower-level reducers and this higher-level function.
        If a hint has already been overridden in the current call stack rooted
        at the top-most call to this function, this function avoids dangerously
        re-overriding nested instances of the same hint. Defaults to the empty
        set.
    pith_name : Optional[str], optional
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    recursable_hint_ids : FrozenSetInts, optional
        **Recursion guard** (i.e., frozen set of the integers uniquely
        identifying all transitive recursable parent hints (i.e., supporting
        recursion) of this hint). If the integer identifying a subsequently
        visited child hint subscripting this hint already resides in this
        recursion guard, that child hint has already been visited by prior
        iteration and is thus a recursive hint. Since recursive hints are valid
        (rather than constituting an unexpected error), the caller is expected
        to detect this use case and silently short-circuit infinite recursion by
        avoiding revisiting that previously visited recursive child hint.
        Defaults to the empty frozen set.
    reductions_count : int, optional
        Current number of total reductions internally performed by *all* calls
        to this function rooted at this function in the current call stack,
        guarding against accidental infinite recursion between lower-level
        reducers and this higher-level function. Defaults to 0.
    typevar_to_hint : TypeVarToHint, optional
        **Type variable lookup table** (i.e., immutable dictionary mapping from
        the :pep:`484`-compliant **type variables** (i.e.,
        :class:`typing.TypeVar` objects) originally parametrizing the origins of
        all transitive parent hints of this hint to the corresponding child
        hints subscripting these parent hints). Defaults to
        :data:`.FROZENDICT_EMPTY`.
    exception_prefix : str, optional
        Human-readable substring prefixing exception messages raised by this
        reducer. Defaults to the empty string.

    Returns
    -------
    HintOrSanifiedData
        Either:

        * If this hint is reducible, either:

          * If reducing this hint to a lower-level hint generates supplementary
            metadata, that metadata including that lower-level hint.
          * Else, that lower-level hint alone.

        * Else, this hint is irreducible. In this case, this hint unmodified.

    Raises
    ------
    BeartypeDecorHintReduceException
        If the number of total reductions internally performed by the current
        call to this function exceeds the maximum. This exception guards against
        accidental infinite recursion between lower-level PEP-specific reducers
        internally called by this higher-level reducer.
    '''

    # ....................{ PREAMBLE                       }....................
    # If passed *NO* hint overrides, default this set to the empty set.
    if hints_overridden is None:
        hints_overridden = set()
    # Else, hint overrides were passed. Preserve these hint overrides as is.

    # Validate passed parameters *AFTER* establishing defaults above.
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'
    assert isinstance(hints_overridden, set), (
        f'{repr(hints_overridden)} not set.')
    assert isinstance(recursable_hint_ids, frozenset), (
        f'{repr(recursable_hint_ids)} not frozen set.')
    assert isinstance(reductions_count, int), (
        f'{repr(reductions_count)} not integer.')
    assert isinstance(typevar_to_hint, FrozenDict), (
        f'{repr(typevar_to_hint)} not frozen dictionary.')

    # ....................{ LOCALS                         }....................
    # Original unreduced hint passed to this reducer, preserved so as to be
    # embedded in human-readable exception messages.
    hint_old = hint

    # Previously reduced instance of this hint, initialized to the sentinel to
    # guarantee that the passed hint is *NEVER* equal to the previously reduced
    # instance of this hint unless actually reduced below. This is necessary, as
    # "None" is a valid type hint reduced to "type(None)" below.
    hint_or_sane_prev: HintOrSanifiedData = SENTINEL  # type: ignore[assignment]

    # BeartypeHintOverrides.get() method bound to the user-defined
    # "hint_overrides" dictionary of this beartype configuration, localized as a
    # negligible efficiency gain.
    conf_hint_overrides_get = conf.hint_overrides.get

    # ....................{ REDUCTION                      }....................
    # Repeatedly reduce this hint to increasingly irreducible hints until this
    # hint is no longer reducible.
    #
    # Note that this algorithm iteratively reduces this hint with a battery of
    # increasingly non-trivial reductions. For efficiency, reductions are
    # intentionally ordered from most to least efficient.
    while True:
        # print(f'[reduce_hint] Reducing {repr(hint)} with type variable lookup table {repr(typevar_to_hint)}...')

        # ....................{ PHASE ~ shallow            }....................
        # Attempt to shallowly reduce this hint to the ignorable "typing.Any"
        # singleton *BEFORE* attempting reductions that deeply inspect the
        # contents of this hint. Why? Several reasons:
        # * This shallow reduction is an O(1) constant-time operation with
        #   negligible constants and thus *INCREDIBLY* fast.
        # * Okay. That's all I've got. It's late, people. I exhaustedly sigh.

        # Machine-readable representation of this hint.
        hint_repr = get_hint_repr(hint)

        #FIXME: Preserved for posterity, as this seems generically useful. *sigh*
        # # If this hint is beartype-blacklisted (i.e., defined in a third-party
        # # package or module that is hostile to runtime type-checking), return true.
        # # print(f'Testing hint {repr(hint)} third-party blacklisting...')
        # if is_object_module_thirdparty_blacklisted(hint):
        #     # print('Blacklisted!')
        #     return True
        # # Else, this hint is *NOT* beartype-blacklisted.

        # If this hint is shallowly ignorable, reduce this hint to the ignorable
        # "typing.Any" singleton.
        #
        # Note that this reduction efficiently applies to multiple signs
        # concurrently, including the "None", "HintSignOptional", and
        # "HintSignUnion". Ergo, this reduction *CANNOT* be trivially
        # implemented as a standard reduction assigned a single sign.
        if hint_repr in HINTS_REPR_IGNORABLE_SHALLOW:
            return AnyObject
        # Else, this hint is *NOT* shallowly ignorable.

        # ....................{ PHASE ~ override           }....................
        # Attempt to reduce this hint to another hint configured by a
        # user-defined hint override *BEFORE* attempting standard reduction.
        # User preference takes precedence over standard precedent.

        #FIXME: *UGH*. There's *NO* need at all for EAFP. Instead:
        #* Refactor this to use hint IDs rather than possibly unhashable hints.
        #* In fact, just refactor this to use our new "recursable_hint_ids"
        #  appoarch entirely, please.

        # Attempt to...
        #
        # Note that the is_object_hashable() tester is internally implemented
        # with the same Easier to Ask for Permission than Forgiveness
        # (EAFP)-based "try-except" block and is thus equally inefficient. In
        # fact, the current approach avoids an extraneous call to that tester
        # and is thus marginally faster. (Emphasis on "marginally.")
        try:
            # If this hint has *NOT* already been overridden by a previously
            # performed reduction in this recursive tree of all previously
            # performed reductions...
            if hint not in hints_overridden:
                # User-defined hint overriding this hint if this beartype
                # configuration overrides this hint *OR* the sentinel otherwise
                # (i.e., if this hist is *NOT* overridden).
                # print(f'Overriding hint {repr(hint)} via {repr(conf.hint_overrides)}...')
                hint_reduced = conf_hint_overrides_get(hint, SENTINEL)

                # If this hint was overridden...
                if hint_reduced is not SENTINEL:
                    # Record this hint as already having been overridden
                    # *BEFORE* reducing (and thus obliterating) this hint.
                    hints_overridden.add(hint)

                    # Reduce this hint to this user-defined hint override.
                    hint = hint_reduced
                # Else, this hint was *NOT* overridden. In this case, preserve
                # this hint as is.
            # Else, this hint has already been overridden by a previously
            # performed reduction in this recursive tree of all previously
            # performed reductions. In this case, avoid overriding this hint yet
            # again. Doing so would (almost certainly) provoke infinite
            # recursion (e.g., overriding "float" with "int | float").
        # If doing so raises a "TypeError", this hint is unhashable and thus
        # inapplicable for hint overriding. In this case, ignore this hint.
        except TypeError:
            pass

        # ....................{ PHASE ~ uncached           }....................
        # Attempt to deeply reduce this possibly contextual hint to another hint
        # *BEFORE* attempting to context-free reductions. Why? Because the
        # former depends on context and is thus "lower-level" than the latter,
        # which depends on *NO* context and is thus "higher-level" in a sense.
        #
        # In theory, order of reduction *SHOULD* be insignificant; in practice,
        # we suspect unforeseen and unpredictable interactions between these two
        # reductions. To reduce the likelihood of fire-breathing dragons here,
        # we reduce contextual hints first.
        hint = _reduce_hint_uncached(
            hint=hint,
            arg_kind=arg_kind,
            cls_stack=cls_stack,
            conf=conf,
            decor_meta=decor_meta,
            hints_overridden=hints_overridden,
            pith_name=pith_name,
            recursable_hint_ids=recursable_hint_ids,
            reductions_count=reductions_count,
            typevar_to_hint=typevar_to_hint,
            exception_prefix=exception_prefix,
        )

        # If this hint reduces to the PEP 484-compliant "typing.Any" singleton,
        # this hint is ignorable. In this case, halt reducing.
        #
        # Note that this is an optional short-circuiting optimization avoiding
        # multiple repetitious reductions for each ignorable hint.
        if hint is Any:
            return AnyObject
        # Else, this hint is currently unignorable. Continue reducing.

        # ....................{ PHASE ~ cached             }....................
        # Sane hint reduced from this possibly insane hint if reducing this hint
        # did not generate supplementary metadata *OR* that metadata otherwise
        # (i.e., if reducing this hint generated supplementary metadata).
        hint_or_sane = _reduce_hint_cached(hint, conf, exception_prefix)

        # If this hint reduces to the PEP 484-compliant "typing.Any" singleton,
        # this hint is ignorable. In this case, halt reducing.
        #
        # Note that, unlike the similar test above, this test is required rather
        # than merely an optimization.
        if hint_or_sane is Any:
            return AnyObject
        # Else, this hint is currently unignorable. Continue reducing.
        #
        # If the current and previously reduced instances of this hint are
        # identical, the above reductions all preserved this hint as is rather
        # than reducing this hint. This implies this hint to now be irreducible.
        # Halt reducing.
        elif hint_or_sane is hint_or_sane_prev:
            break
        # Else, the current and previously reduced instances of this hint
        # differ, implying this hint to still be reducible. Continue reducing.

        # This possibly context-free hint efficiently reduced to another hint
        # and the resulting type variable lookup table.
        hint, typevar_to_hint = unpack_hint_or_sane(
            hint_or_sane=hint_or_sane,
            typevar_to_hint=typevar_to_hint,
        )
        # print(f'[reduce_hint] Reduced to {repr(hint)} and type variable lookup table {repr(typevar_to_hint)}.')

        # Increment the current number of total reductions internally performed
        # by this call *BEFORE* detecting accidental recursion below.
        reductions_count += 1

        # If the current number of total reductions internally performed
        # by this call exceeds the maximum, raise an exception.
        #
        # Note that this should *NEVER* happen, but probably nonetheless will.
        if reductions_count >= _REDUCTIONS_COUNT_MAX:  # pragma: no cover
            raise BeartypeDecorHintReduceException(
                f'{exception_prefix}type hint {repr(hint_old)} irreducible; '
                f'recursion detected when reducing between reduced type hints '
                f'{repr(hint_or_sane_prev)} and {repr(hint)}. '
                f'Please submit this exception traceback as a new issue '
                f'on our friendly issue tracker:\n'
                f'\t{URL_ISSUES}\n'
                f'Beartype thanks you for your noble (yet ultimately tragic) '
                f'sacrifice.'
            )
        # Else, the current number of total reductions internally performed
        # by this call is still less than the maximum. In this case, continue.

        # Previously reduced instance of this hint.
        hint_or_sane_prev = hint

    # ....................{ RETURN                         }....................
    # Either this possibly reduced hint *OR* metadata describing this hint,
    # defined as either...
    hint_or_sane = (
        # If either...
        #
        # Then *NO* metadata describes this reduction. In this case, ignore this
        # metadata by reducing to this hint alone.
        hint
        if (
            #FIXME: Maybe *UNTRUE* if "recursable_hint_ids" is non-empty?
            # This hint reduces to the PEP 484-compliant "typing.Any" singleton,
            # this hint is ignorable. Since this hint is ignorable, any
            # extraneous metadata associated with this hint (e.g., type variable
            # lookup table) is also ignorable. We intentionally ignore this
            # metadata by reducing this hint to simply "typing.Any",
            # trivializing detection of ignorable hints throughout the codebase;
            hint is Any or
            # This hint maps *NO* type variables.
            not typevar_to_hint
        ) else
        # Else, this reduction maps one or more type variables. In this case,
        # metadata describing both this hint and this mapping.
        HintSanifiedData(hint=hint, typevar_to_hint=typevar_to_hint)
    )

    # Return this possibly reduced hint.
    return hint_or_sane

# ....................{ PRIVATE ~ reducers                 }....................
# Note that the higher-level reduce_hint() reducer defined above sequentially
# calls the lower-level _reduce_hint_uncached() reducer defined below *BEFORE*
# calling the lower-level _reduce_hint_cached() reducer defined below. As such,
# the current hint should be validated in the former rather than the latter.

def _reduce_hint_uncached(
    hint: Hint,
    arg_kind: Optional[ArgKind],
    cls_stack: TypeStack,
    conf: BeartypeConf,
    decor_meta: Optional[BeartypeDecorMeta],
    hints_overridden: SetHints,
    pith_name: Optional[str],
    recursable_hint_ids: FrozenSetInts,
    reductions_count: int,
    typevar_to_hint: TypeVarToHint,
    exception_prefix: str,
) -> Hint:
    '''
    Lower-level **contextual type hint** (i.e., type hint contextually dependent
    on the kind of class, attribute, callable parameter, or callable return
    annotated by this hint) inefficiently reduced (i.e., converted) from the
    passed higher-level context-free type hint if this hint is reducible *or*
    this hint as is otherwise (i.e., if this hint is irreducible).

    This reducer *cannot* be meaningfully memoized, since multiple passed
    parameters (e.g., ``pith_name``, ``cls_stack``) are typically isolated to a
    handful of callables across the codebase currently being decorated by
    :mod:`beartype`. Thankfully, this reducer is responsible for reducing only a
    small subset of type hints requiring these problematic parameters.

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.
    cls_stack : TypeStack
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    decor_meta : Optional[BeartypeDecorMeta]
        Either:

        * If this hint annotates a parameter or return of some callable, the
          :mod:`beartype`-specific decorator metadata describing that callable.
        * Else, :data:`None`.
    hints_overridden : set[Hint]
        Mutable set of all previously **overridden hints** (i.e., hints reduced
        to user-defined hint overrides configured by the
        :attr:`beartype.BeartypeConf.hint_overrides` dictionary) reduced by a
        prior call to this function, guarding against accidental infinite
        recursion between lower-level reducers and this higher-level function.
        If a hint has already been overridden in the current call stack rooted
        at the top-most call to this function, this function avoids dangerously
        re-overriding nested instances of the same hint.
    pith_name : Optional[str]
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.
    recursable_hint_ids : FrozenSetInts
        **Recursion guard** (i.e., frozen set of the integers uniquely
        identifying all transitive recursable parent hints (i.e., supporting
        recursion) of this hint).
    reductions_count : int
        Current number of total reductions internally performed by *all* calls
        to this function rooted at this function in the current call stack,
        guarding against accidental infinite recursion between lower-level
        reducers and this higher-level function.
    typevar_to_hint : TypeVarToHint
        **Type variable lookup table** (i.e., immutable dictionary mapping from
        the :pep:`484`-compliant **type variables** (i.e.,
        :class:`typing.TypeVar` objects) originally parametrizing the origins of
        all transitive parent hints of this hint to the corresponding child
        hints subscripting these parent hints).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        Either:

        * If the passed hint is reducible, another hint reduced from this hint.
        * Else, this hint as is unmodified.
    '''

    # Sign uniquely identifying this hint if this hint is PEP-compliant *OR*
    # "None" otherwise (e.g., if this hint is PEP-noncompliant).
    hint_sign = get_hint_pep_sign_or_none(hint)
    # print(f'_reduce_hint_uncached() hint_sign: {hint_sign}')

    # If this hint is PEP-noncompliant...
    #
    # Note that this logic could also be handled in a reduce_hint_nonpep()
    # reducer mapped to the "None" sign, whose body is the body of this "if"
    # conditional. Indeed, the prior implementation of this logic did just that.
    # Technically, that approach did work. Pragmatically, that approach was
    # obfuscatory, obtuse, and ultimately unmaintainable. This approach is
    # equally efficient but *CONSIDERABLY* more sensible. Sanity wins.
    if hint_sign is None:
        # If this hint is unsupported by @beartype, raise an exception.
        die_unless_hint(hint=hint, exception_prefix=exception_prefix)
        # Else, this hint is supported by @beartype.

        # Return this hint as is. By definition, PEP-noncompliant hints are
        # irreducible. If this hint was instead reducible, the
        # get_hint_pep_sign_or_none() getter called above would have instead
        # returned a unique sign identifying this hint (rather than "None").
        return hint
    # Else, this hint is PEP-compliant.

    # Callable reducing this hint if a callable reducing hints of this sign was
    # previously registered *OR* "None" otherwise (i.e., if *NO* such callable
    # was registered, in which case this hint is preserved as is).
    hint_reducer = HINT_SIGN_TO_REDUCE_HINT_UNCACHED_get(hint_sign)

    # If a callable reducing hints of this sign was previously registered,
    # reduce this hint to another hint via this callable.
    if hint_reducer is not None:  # type: ignore[call-arg]
        # print(f'[_reduce_hint_uncached] Reducing uncached hint {repr(hint)}...')
        # print(f'...with type variable lookup table {repr(typevar_to_hint)}...')
        hint = hint_reducer(
            hint=hint,  # pyright: ignore
            arg_kind=arg_kind,
            cls_stack=cls_stack,
            conf=conf,
            decor_meta=decor_meta,
            hints_overridden=hints_overridden,
            pith_name=pith_name,
            recursable_hint_ids=recursable_hint_ids,
            reductions_count=reductions_count,
            typevar_to_hint=typevar_to_hint,
            exception_prefix=exception_prefix,
        )
        # print(f'[_reduce_hint_uncached]...to uncached hint {repr(hint)}.')
    # Else, *NO* such callable was registered. Preserve this hint as is, you!

    # Return this possibly reduced hint.
    return hint


@callable_cached
def _reduce_hint_cached(
    hint: Hint,
    conf: BeartypeConf,
    exception_prefix: str,
) -> HintOrSanifiedData:
    '''
    Lower-level **context-free type hint** (i.e., type hint *not* contextually
    dependent on the kind of class, attribute, callable parameter, or callable
    return annotated by this hint) efficiently reduced (i.e., converted) from
    the passed higher-level context-free type hint if this hint is reducible
    *or* this hint as is otherwise (i.e., if this hint is irreducible).

    This reducer is memoized for efficiency. Thankfully, this reducer is
    responsible for reducing *most* (but not all) type hints.

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    exception_prefix : str
        Substring prefixing exception messages raised by this function.

    Returns
    -------
    HintOrSanifiedData
        Either:

        * If the passed hint is reducible, either:

          * If reducing this hint to a lower-level hint generates supplementary
            metadata, that metadata including that lower-level hint.
          * Else, that lower-level hint alone.

        * Else, this hint is irreducible. In this case, this hint unmodified.
    '''

    # Sign uniquely identifying this hint if this hint is identifiable *OR*
    # "None" otherwise (e.g., if this hint is merely an isinstanceable class).
    hint_sign = get_hint_pep_sign_or_none(hint)

    # Callable reducing this hint if a callable reducing hints of this sign was
    # previously registered *OR* "None" otherwise (i.e., if *NO* such callable
    # was registered, in which case this hint is preserved as is).
    hint_reducer = HINT_SIGN_TO_REDUCE_HINT_CACHED_get(hint_sign)

    # If a callable reducing hints of this sign was previously registered,
    # reduce this hint to another hint via this callable.
    if hint_reducer is not None:
        # print(f'[_reduce_hint_cached] Reducing cached hint {repr(hint)}...')
        hint = hint_reducer(  # type: ignore[call-arg]
            hint=hint,  # pyright: ignore[reportGeneralTypeIssues]
            conf=conf,
            exception_prefix=exception_prefix,
        )
        # print(f'[_reduce_hint_cached] ...to cached hint {repr(hint)}.')
    # Else, *NO* such callable was registered. Preserve this hint as is, you!

    # Return this possibly reduced hint.
    return hint

# ....................{ PRIVATE ~ globals                  }....................
_REDUCTIONS_COUNT_MAX = 64
'''
Maximum number of total reductions internally performed by each call of the
:func:`reduce_hint` function.

This constant is a relatively arbitrary magic number selected so as to guard
against accidental infinite recursion between lower-level PEP-specific reducers
internally called by :func:`reduce_hint`.
'''


_REDUCE_HINT_CHILD_ARG_NAMES_UNSAFE = frozenset((
    # Applicable *ONLY* to root hints directly annotating callable parameters.
    'arg_kind',

    # Applicable *ONLY* to root hints directly annotating callable parameters
    # or returns.
    'decor_meta',
    'pith_name',
))
'''
Frozen set of the names of all **unsafe child type hint reducer keyword
parameters** (i.e., keyword parameters inapplicable to child type hints and thus
*not* safely passable from the higher-level :func:`.reduce_hint_child` to
lower-level :func:`.reduce_hint` reducer.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695` **type alias data submodule.**

This submodule defines :pep:`695`-compliant type aliases to be exercised
throughout unit tests exercising these aliases. For safety, these aliases are
intentionally isolated from those unit tests.

Motivation
----------
Generic type alias support is unavoidably implemented in an *extremely* fragile
manner throughout the :mod:`beartype` codebase. Recursively "bubbling up" the
concrete child type hints subscripting a generic type alias into the lower-level
target type hint aliased by that generic type alias is so non-trivial that it
necessitates implementing manual support for "bubbling up" across both all
hint-specific dynamic code generators *and* all hint-specific dynamic exception
raisers: which is to say, the *entire* :mod:`beartype._check` subpackage.

Generic type alias support is so widely distributed that the *only* valid
means of testing this support is to exhaustively define here and subsequently
use below at least one generic type alias for *each* other PEP standard
supported by :mod:`beartype`. Non-trivial, thy name is :pep:`695`.

Caveats
-------
**This submodule requires the active Python interpreter to target at least
Python 3.12.0.** If this is *not* the case, importing this submodule raises an
:exc:`SyntaxError` exception.
'''

# ..................{ IMPORTS                                }..................
from beartype.typing import (
    Annotated,
    Union,
)
from beartype.vale import Is
from beartype_test.a00_unit.data.hint.pep.proposal.pep484585.data_pep484585generic import (
    Pep585IterableTContainerT,
    Pep585IterableTupleSTContainerTupleST,
)

# ..................{ SIMPLE                                 }..................
type AliasSimple = int | list[str]
'''
Simple type alias aliasing a standard type hint containing *NO* syntax or
semantics unique to PEP 695-compliant type aliases (e.g., *NO* forward
references, recursion, or type variables).
'''

# ..................{ PEP (484|604)                          }..................
type AliasPep484604[T] = (
    # PEP-noncompliant type.
    float |
    # PEP 604-compliant new union of two or more generic type hints
    # parametrized by the same type variable subscripting this alias.
    list[T] | set[T] |
    # PEP 484-compliant old union of (...waitforit) two or more generic type
    # hints parametrized by the same type variable subscripting this alias.
    Union[dict[T, T] | frozenset[T]]
)
'''
Type alias aliasing a :pep:`604`-compliant new union of:

* A PEP-noncompliant type.
* Two or more generic type hints parametrized by the same type variable
  subscripting this alias.
* A :pep:`484`-compliant old union of two or more generic type hints
  parametrized by the same type variable subscripting this alias.

This alias thus tests that dynamic code generation for both :pep:`484`- and
:pep:`604`-compliant unions preserves :pep:`484`-compliant type variable
mappings while also flattening directly nested unions into the top-level union.
'''

# ..................{ PEP (484|604) ~ depth ~ unparametrized }..................
type AliasPep484604Depth1 = _AliasPep484604Depth2 | float
'''
Type alias aliasing a :pep:`604`-compliant new union of a PEP-noncompliant type
and another type alias aliasing a :pep:`484`-compliant old union of
another PEP-noncompliant type and another type alias aliasing (...waitforit)
another PEP-noncompliant type.

This alias thus tests that dynamic code generation for both :pep:`484`- and
:pep:`604`-compliant unions recursively unflattens deeply nested unions into the
top-level union.
'''


type _AliasPep484604Depth2 = Union[str, _AliasPep484604Depth3]
'''
Type alias aliasing a :pep:`484`-compliant old union of a PEP-noncompliant type
and another type alias aliasing another PEP-noncompliant type.
'''


type _AliasPep484604Depth3 = int
'''
Type alias aliasing a PEP-noncompliant type.
'''

# ..................{ PEP (484|604) ~ depth ~ parametrized   }..................
type AliasPep484604Depth1T[T] = _AliasPep484604Depth2T[T] | float
'''
Type alias aliasing a :pep:`604`-compliant new union of a PEP-noncompliant type
and another type alias subscripted by the same type variable parametrizing this
alias aliasing a :pep:`484`-compliant old union of another PEP-noncompliant type
and another type alias subscripted by the same type variable parametrizing this
alias aliasing that type parameter.

This alias thus tests that dynamic code generation for both :pep:`484`- and
:pep:`604`-compliant unions recursively unflattens deeply nested unions into the
top-level union.
'''


type _AliasPep484604Depth2T[T] = Union[str, _AliasPep484604Depth3T[T]]
'''
Type alias aliasing a :pep:`484`-compliant old union of a PEP-noncompliant type
and another type alias subscripted by the same type variable parametrizing this
alias aliasing another PEP-noncompliant type .
'''


type _AliasPep484604Depth3T[T] = T
'''
**Identity type alias** isomorphically aliasing the type parameter parametrizing
this alias.
'''

# ..................{ PEP 585                                }..................
type AliasPep585Dict[S, T] = dict[S, T]
'''
Type alias aliasing a :pep:`585`-compliant dictionary subscripted by the same
type variables as those parametrizing this alias.
'''


type AliasPep585TupleFixed[S, T] = tuple[S, T]
'''
Type alias aliasing a :pep:`585`-compliant fixed-length tuple type hint
subscripted by the same type variables as those parametrizing this alias.
'''


type AliasPep585Type[T] = type[T]
'''
Type alias aliasing a :pep:`585`-compliant subclass type hint subscripted by the
same type variable as that parametrizing this alias.
'''

# ..................{ PEP 585 ~ generic                      }..................
type AliasPep585IterableTupleSTContainerTupleST[S, T] = (
    Pep585IterableTupleSTContainerTupleST[S, T])
'''
Type alias aliasing a :pep:`585`-compliant generic subscripted by the same type
variables as those parametrizing this alias.
'''


type AliasPep585IterableTContainerT[T] = T | Pep585IterableTContainerT[T] | None
'''
Type alias aliasing a :pep:`604`-compliant union over a type variable, a
:pep:`585`-compliant generic subscripted by that same type variable, and
:data:`None`.

This alias intentionally exercises an edge case that previously induced infinite
recursion. Badness lurks below.
'''

# ..................{ PEP 593                                }..................
type AliasPep593[T] = Annotated[T | bytes, Is[lambda obj: bool(obj)]]
'''
Type alias aliasing a PEP 593-compliant metahint subscripted by the same
type variable as that parametrizing this alias.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype metadata.**

This submodule exports global constants synopsizing this package -- including
versioning and dependencies.

Python Version
--------------
For uniformity between this package and the ``pyproject.toml`` file describing
the installation of this package, this submodule also validates the version of
the active Python interpreter. An exception is raised if this version is
insufficient.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: To avoid polluting the public module namespace, external attributes
# should be locally imported at module scope *ONLY* under alternate private
# names (e.g., "from argparse import ArgumentParser as _ArgumentParser" rather
# than merely "from argparse import ArgumentParser").
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys as _sys
from beartype._util.text.utiltextversion import (
    convert_str_version_to_tuple as _convert_str_version_to_tuple)
from importlib.metadata import metadata as _get_package_metadata
from typing import (
    TYPE_CHECKING,  # <-- *MUST* be import as "TYPE_CHECKING" or mypy ignores it
    Optional as _Optional,
)

# ....................{ METADATA                           }....................
NAME = 'beartype'
'''
Human-readable package name.
'''


# Ideally, this metadata would be parsed from the "_package_metadata" dictionary
# introspected below. Sadly, this metadata has yet to be standardized. The
# closest approximation is the "_package_metadata['License']" key, which
# provides the contents of the top-level "LICENSE" file rather than the name of
# the license licensing this package. It is what it is. It is sucky. *sigh*
LICENSE = 'MIT'
'''
Human-readable name of the license this package is licensed under.
'''

# ....................{ METADATA                           }....................
# If performing static type-checking, define a fake "_package_metadata"
# dictionary as a crude means of informing the static type-checker of the
# expected type of this dictionary. While awful, this is (probably) the least
# awful approach. All alternatives invite deprecation concerns.
if TYPE_CHECKING:
    _package_metadata = {'Requires-Python': '>=3.9'}
# Else, static type-checking is *NOT* being performed. In this case...
else:
    # Dictionary mapping from the name to value of all core packaging metadata
    # with which this package was installed under the active Python interpreter.
    #
    # See also the "Core metadata specifications," which standardizes the names
    # and values of this metadata via the PEP process:
    #     https://packaging.python.org/en/latest/specifications/core-metadata
    #
    # First, attempt to introspect this metadata from the physical (i.e.,
    # on-disk) distribution describing this package.
    try:
        _package_metadata = _get_package_metadata(NAME)
    # If doing so fails for *ANY* reason whatsoever, silently ignore this
    # failure by falling back to a default dictionary permissively mapping the
    # names of this metadata to the placeholder "None".
    #
    # Note that this edge case occurs in common use cases that compile,
    # transpile, or freeze this package. Downstream consumers of this submodule
    # *MUST* thus explicitly detect imported globals whose values are "None" and
    # react nicely.
    except:
        from collections import defaultdict as _defaultdict
        _package_metadata = _defaultdict(lambda: None)

# ....................{ METADATA ~ license                 }....................
PACKAGE_NAME = NAME
'''
Fully-qualified name of the top-level Python package containing this submodule.
'''


PACKAGE_TEST_NAME = f'{PACKAGE_NAME}_test'
'''
Fully-qualified name of the top-level Python package exercising this project.
'''

# ....................{ PYTHON ~ version                   }....................
PYTHON_VERSION_MIN: _Optional[str] = (
    # If this package distribution defines the "Requires-Python" key, the value
    # of this key stripped of its ">=" prefix. Notably, the value of this key is
    # the value of the "requires-python" key in the "pyproject.toml" file: e.g.,
    #     requires-python = ">=3.8"
    #
    # Since the latter is guaranteed to be prefixed by the substring ">=" of
    # length 2, removing this prefix from this string yields the minimum version
    # of Python required by this package as a "."-delimited string. Phew!
    _package_metadata['Requires-Python'][2:]
    if _package_metadata['Requires-Python'] else
    # Else, this package distribution fails to define this key. In this case,
    # fallback to "None".
    None
)
'''
Human-readable minimum version of Python required by this package as a
``.``-delimited string if this package distribution provides this metadata *or*
:data:`None` otherwise (i.e., if this package distribution fails to provide this
metadata).
'''


PYTHON_VERSION_MIN_PARTS = (
    # If this package distribution defines the "Requires-Python" key, the value
    # of this key stripped of its ">=" prefix and coerced into a tuple of
    # integers.
    _convert_str_version_to_tuple(PYTHON_VERSION_MIN)
    if PYTHON_VERSION_MIN is not None else
    # Else, this package distribution fails to define this key. In this case,
    # fallback to "None".
    None
)
'''
Machine-readable minimum version of Python required by this package as a
tuple of integers if this package distribution provides this metadata *or*
:data:`None` otherwise (i.e., if this package distribution fails to provide this
metadata).
'''

# ....................{ METADATA ~ version                 }....................
VERSION = '0.20.2'
'''
Human-readable package version as a ``.``-delimited string.
'''


VERSION_PARTS = _convert_str_version_to_tuple(VERSION)
'''
Machine-readable package version as a tuple of integers.
'''

# ....................{ METADATA ~ synopsis                }....................
SYNOPSIS: _Optional[str] = _package_metadata['Summary']
'''
Human-readable single-line synopsis of this package.

By PyPI design, this string must *not* span multiple lines or paragraphs.
'''

# ....................{ METADATA ~ authors                 }....................
AUTHOR_EMAIL: _Optional[str] = _package_metadata['Author-email']
'''
Email address of the principal corresponding author (i.e., the principal author
responding to public correspondence).
'''


AUTHORS = 'Cecil Curry, et al.'
'''
Human-readable list of all principal authors of this package as a
comma-delimited string.

For brevity, this string *only* lists authors explicitly assigned copyrights.
For the list of all contributors regardless of copyright assignment or
attribution, see the `contributors graph`_ for this project.

.. _contributors graph:
   https://github.com/beartype/beartype/graphs/contributors
'''


COPYRIGHT = '2014-2025 Beartype authors'
'''
Legally binding copyright line excluding the license-specific prefix (e.g.,
``"Copyright (c)"``).

For brevity, this string *only* lists authors explicitly assigned copyrights.
For the list of all contributors regardless of copyright assignment or
attribution, see the `contributors graph`_ for this project.

.. _contributors graph:
   https://github.com/beartype/beartype/graphs/contributors
'''

# ....................{ METADATA ~ urls                    }....................
# Although feasible, parsing URLs from "_package_metadata" is non-trivial.
# Rather than break our body over something nobody cares about, violate the DRY
# (Don't Repeat Yourself) principle by repeating various URLs already specified
# in our top-level "pyproject.toml" file.

URL_CONDA = f'https://anaconda.org/conda-forge/{PACKAGE_NAME}'
'''
URL of this project's entry on **Anaconda** (i.e., alternate third-party Python
package repository utilized by the Anaconda Python distribution).
'''


URL_LIBRARIES = f'https://libraries.io/pypi/{PACKAGE_NAME}'
'''
URL of this project's entry on **Libraries.io** (i.e., third-party open-source
package registrar associated with the Tidelift open-source funding agency).
'''


URL_PYPI = f'https://pypi.org/project/{PACKAGE_NAME}'
'''
URL of this project's entry on **PyPI** (i.e., official Python package
repository, also colloquially known as the "cheeseshop").
'''


URL_RTD = f'https://readthedocs.org/projects/{PACKAGE_NAME}'
'''
URL of this project's entry on **ReadTheDocs (RTD)** (i.e., popular Python
documentation host, shockingly hosting this project's documentation).
'''

# ....................{ METADATA ~ urls : docs             }....................
URL_HOMEPAGE = f'https://{PACKAGE_NAME}.readthedocs.io'
'''
URL of this project's homepage.
'''


URL_PEP585_DEPRECATIONS = (
    f'{URL_HOMEPAGE}/en/latest/api_roar/#pep-585-deprecations')
'''
URL documenting :pep:`585` deprecations of :pep:`484` type hints.
'''

# ....................{ METADATA ~ urls : repo             }....................
URL_REPO_ORG_NAME = PACKAGE_NAME
'''
Name of the **organization** (i.e., parent group or user principally responsible
for maintaining this project, indicated as the second-to-last trailing
subdirectory component) of the URL of this project's git repository.
'''


URL_REPO_BASENAME = PACKAGE_NAME
'''
**Basename** (i.e., trailing subdirectory component) of the URL of this
project's git repository.
'''


URL_REPO = f'https://github.com/{URL_REPO_ORG_NAME}/{URL_REPO_BASENAME}'
'''
URL of this project's git repository.
'''


URL_DOWNLOAD = f'{URL_REPO}/archive/{VERSION}.tar.gz'
'''
URL of the source tarball for the current version of this project.

This URL assumes a tag whose name is ``v{VERSION}`` where ``{VERSION}`` is the
human-readable current version of this project (e.g., ``v0.4.0``) to exist.
Typically, no such tag exists for live versions of this project -- which
have yet to be stabilized and hence tagged. Hence, this URL is typically valid
*only* for previously released (rather than live) versions of this project.
'''


URL_FORUMS = f'{URL_REPO}/discussions'
'''
URL of this project's user forums.
'''


URL_ISSUES = f'{URL_REPO}/issues'
'''
URL of this project's issue tracker.
'''


URL_RELEASES = f'{URL_REPO}/releases'
'''
URL of this project's release list.
'''

# ....................{ METADATA ~ dependency : names      }....................
#FIXME: Switch! So, "pydata-sphinx-theme" is ostensibly *MOSTLY* great. However,
#there are numerous obvious eccentricities in "pydata-sphinx-theme" that we
#strongly disagree with -- especially that theme's oddball division in TOC
#heading levels between the top and left sidebars.
#
#Enter "sphinx-book-theme", stage left. "sphinx-book-theme" is based on
#"pydata-sphinx-theme", but entirely dispenses with all of the obvious
#eccentricities that hamper usage of "pydata-sphinx-theme". We no longer have
#adequate time to maintain custom documentation CSS against the moving target
#that is "pydata-sphinx-theme". Ergo, we should instead let "sphinx-book-theme"
#do all of that heavy lifting for us. Doing so will enable us to:
#* Lift the horrifying constraint above on a maximum Sphinx version. *gulp*
#* Substantially simplify our Sphinx configuration. Notably, the entire fragile
#  "doc/src/_templates/" subdirectory should be *ENTIRELY* excised away.
#
#Please transition to "sphinx-book-theme" as time permits.

# Note that documentation-time functionality in the Sphinx-specific
# "doc/src/conf.py" script imports this private string global. *shrug*
SPHINX_THEME_NAME = 'pydata-sphinx-theme'
'''
Name of the third-party Sphinx extension providing the custom HTML theme
preferred by this documentation.

See Also
--------
pyproject.toml
    Further discussion in the ``doc-rtd`` key of our top-level
    ``pyproject.toml`` file.
'''

# ....................{ METADATA ~ dependency : versions   }....................
# Note that test-time functionality imports this private string global. *shrug*
_LIB_RUNTIME_OPTIONAL_VERSION_MINIMUM_NUMPY = '1.21.0'
'''
Minimum optional version of NumPy recommended for use with this project.

NumPy >= 1.21.0 first introduced the third-party PEP-noncompliant
:attr:`numpy.typing.NDArray` type hint supported by the
:func:`beartype.beartype` decorator.
'''


# Note that test-time functionality imports this private string global. *shrug*
_LIB_RUNTIME_OPTIONAL_VERSION_MINIMUM_TYPING_EXTENSIONS = '3.10.0.0'
'''
Minimum optional version of the third-party :mod:`typing_extensions` package
recommended for use with this project.

:mod:`typing_extensions` >= 3.10.0.0 backports all :mod:`typing` attributes
unavailable under older Python interpreters supported by the
:func:`beartype.beartype` decorator.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **subclass type hint
utilities** (i.e., callables generically applicable to both :pep:`484`- and
:pep:`585`-compliant subclass type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.hint.datahintpep import Hint
from typing import (
    Any,
    Type as typing_Type,  # <-- intentional to distinguish from "type" below
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **type-checking code container classes** (i.e., low-level classes
storing metadata describing the breadth-first search (BFS) dynamically
generating pure-Python code snippets type-checking arbitrary objects against
PEP-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    TYPE_CHECKING,
    # Any,
    Optional,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.code.codemagic import (
    EXCEPTION_PREFIX_FUNC_WRAPPER_LOCAL)
from beartype._check.code.codescope import add_func_scope_type_or_types
from beartype._check.code.snip.codesnipcls import (
    PITH_INDEX_TO_VAR_NAME)
from beartype._check.convert.convsanify import sanify_hint_child
from beartype._check.metadata.hint.hintmeta import HintMeta
from beartype._check.metadata.metasane import (
    HintOrSanifiedData,
    unpack_hint_or_sane,
)
from beartype._conf.confcls import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.code.datacodeindent import INDENT_LEVEL_TO_CODE
from beartype._data.error.dataerrmagic import (
    EXCEPTION_PLACEHOLDER as EXCEPTION_PREFIX)
from beartype._data.hint.datahintpep import (
    # ANY,
    Hint,
    TypeVarToHint,
)
from beartype._data.hint.datahinttyping import (
    FrozenSetInts,
    LexicalScope,
    TypeStack,
    TypeOrSetOrTupleTypes,
)
# from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._data.kind.datakindset import FROZENSET_EMPTY
from beartype._util.cache.pool.utilcachepoollistfixed import (
    FIXED_LIST_SIZE_MEDIUM,
    FixedList,
)
from beartype._util.kind.map.utilmapfrozen import FrozenDict

# ....................{ SUBCLASSES                         }....................
#FIXME: Unit test us up, please.
class HintsMeta(FixedList):
    '''
    **Type hint type-checking metadata queue** (i.e., low-level fixed list of
    metadata describing all visitable type hints currently discovered by the
    breadth-first search (BFS) dynamically generating pure-Python type-checking
    code snippets in the :func:`beartype._check.code.codemake.make_check_expr`
    factory).

    This list acts as a standard First In First Out (FILO) queue, enabling that
    BFS to be implemented as an efficient imperative algorithm rather than an
    inefficient -- and dangerous, due to both unavoidable stack exhaustion and
    avoidable infinite recursion -- recursive algorithm.

    Note that this list is guaranteed by the previously called
    ``_die_if_hint_repr_exceeds_child_limit()`` function to be larger than the
    number of hints transitively visitable from this root hint. Ergo, *all*
    indexation into this list performed by this BFS is guaranteed to be safe.

    Design
    ------
    Most of the following instance variables are only relevant when the
    currently visited hint is *not* the root hint. If the currently visited hint
    is the root hint, the current pith has already been localized to a local
    variable whose name is the value of the :data:`VAR_NAME_PITH_ROOT` string
    global and thus need *not* be relocalized to another local variable using an
    assignment expression.

    These variables enable a non-trivial runtime optimization eliminating
    repeated computations to obtain the child pith needed to type-check child
    hints. For example, if the current hint constrains the current pith to be
    a standard sequence, the child pith of that parent pith is a random item
    selected from this sequence; since obtaining this child pith is
    non-trivial, the computation required to do so is performed only once by
    assigning this child pith to a unique local variable during type-checking
    and then repeatedly type-checking that variable rather than the logic
    required to continually reacquire this child pith: e.g.,

    .. code-block:: python

       # Type-checking conditional for "List[List[str]]" under Python < 3.8.
       if not (
           isinstance(__beartype_pith_0, list) and
           (
               isinstance(__beartype_pith_0[__beartype_random_int % len(__beartype_pith_0)], list) and
               isinstance(__beartype_pith_0[__beartype_random_int % len(__beartype_pith_0)][__beartype_random_int % len(__beartype_pith_0[__beartype_random_int % len(__beartype_pith_0)])], str) if __beartype_pith_0[__beartype_random_int % len(__beartype_pith_0)] else True
           ) if __beartype_pith_0 else True
       ):

       # The same conditional under Python >= 3.8.
       if not (
           isinstance(__beartype_pith_0, list) and
           (
               isinstance(__beartype_pith_1 := __beartype_pith_0[__beartype_random_int % len(__beartype_pith_0)], list) and
               isinstance(__beartype_pith_1[__beartype_random_int % len(__beartype_pith_1)], str) if __beartype_pith_1 else True
           ) if __beartype_pith_0 else True
       ):

    Note that:

    * The random item selected from the root pith (i.e., ``__beartype_pith_1
      := __beartype_pith_0[__beartype_random_int % len(__beartype_pith_0)``)
      only occurs once under Python >= 3.8 but repeatedly under Python < 3.8.
      In both cases, the same semantic type-checking is performed regardless
      of optimization.
    * This optimization implicitly "bottoms out" when the currently visited hint
      is *not* subscripted by unignorable child hints. If all child hints of the
      currently visited hint are either ignorable (e.g., :class:`object`,
      :obj:`typing.Any`) *or* are unignorable isinstanceable types (e.g.,
      :class:`int`, :class:`str`), the currently visited hint has *no*
      meaningful child hints and is thus effectively a leaf node with respect to
      performing this optimization.

    Attributes
    ----------
    cls_stack : TypeStack
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.
    func_curr_code : Optional[str]
        Either:

        * If the currently visited hint is deeply type-checkable, the Python
          code snippet type-checking the current pith against this hint.
        * If the currently visited hint is only shallowly type-checkable,
          :data:`None`.
    func_wrapper_scope : LexicalScope
        **Local scope** (i.e., dictionary mapping from the name to value of each
        attribute referenced in the signature) of this wrapper function required
        by this Python code snippet.
    hint_curr_expr : Optional[str]
        Either:

        * If the currently visited hint is deeply type-checkable, :data:`None`.
        * If the currently visited hint is only shallowly type-checkable, the
          Python expression evaluating to the origin type underlying this hint
          as a hidden :mod:`beartype`-specific parameter injected into the
          signature of the current wrapper function.
    hint_curr_meta: HintMeta
        Metadata describing the currently visited hint, appended by the
        previously visited parent hint to this queue.
    indent_curr : str
        Python code snippet expanding to the current level of indentation
        appropriate for the currently visited hint.
    indent_child : str
        Python code snippet expanding to the current level of indentation
        appropriate for the currently iterated child hint of this parent hint.
    indent_level_child : int
        1-based indentation level describing the current level of indentation
        appropriate for the currently iterated child hint of this parent hint.
    index_last : int
        0-based index of metadata describing the last visitable hint in this
        list. For efficiency, this integer also uniquely identifies the current
        child type hint of the currently visited parent type hint.
    is_var_random_int_needed : bool
        :data:`True` only if one or more child hints of the root hint of this
        queue require a pseudo-random integer. If :data:`True`, the body of this
        wrapper function will be prefixed with code generating this integer.
    pith_curr_expr : str
        Full Python expression evaluating to the value of the **current pith**
        (i.e., possibly nested object of the current parameter or return value
        to be type-checked against this union type hint).

        Note that this is intentionally *not* an assignment expression but
        rather the original inefficient expression provided by the parent type
        hint of the currently visited hint.
    pith_curr_assign_expr : str
        Assignment expression assigning this full Python expression to the
        unique local variable assigned the value of this expression.
    pith_curr_var_name : str
        Name of the current pith variable (i.e., local Python variable in the
        body of the wrapper function whose value is that of the current pith).
        This name is either:

        * Initially, the name of the currently type-checked parameter or return.
        * On subsequently type-checking nested items of the parameter or return,
          the name of the local variable uniquely assigned to by the assignment
          expression defined by :attr:`pith_curr_assign_expr` (i.e., the
          left-hand side (LHS) of that assignment expression).
    pith_curr_var_name_index : int
        Integer suffixing the name of each local variable assigned the value of
        the current pith in a assignment expression, thus uniquifying this
        variable in the body of the current wrapper function.

        Note that this integer is intentionally incremented as an efficient
        low-level scalar rather than as an inefficient high-level
        "itertools.Counter" object. Since both are equally thread-safe in the
        internal context of this dataclass, the former is preferable.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'cls_stack',
        'conf',
        'exception_prefix',
        'func_curr_code',
        'func_wrapper_scope',
        'hint_curr_expr',
        'hint_curr_meta',
        'indent_curr',
        'indent_child',
        'indent_level_child',
        'index_last',
        'is_var_random_int_needed',
        'pith_curr_expr',
        'pith_curr_assign_expr',
        'pith_curr_var_name',
        'pith_curr_var_name_index',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        cls_stack: TypeStack
        conf: BeartypeConf
        exception_prefix: str
        func_curr_code: str
        func_wrapper_scope: LexicalScope
        hint_curr_expr : Optional[str]
        hint_curr_meta : HintMeta
        indent_curr: str
        indent_child: str
        indent_level_child: int
        index_last: int
        is_var_random_int_needed: bool
        pith_curr_expr: str
        pith_curr_assign_expr: str
        pith_curr_var_name: str
        pith_curr_var_name_index: int

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this type-checking metadata list.
        '''

        # Initialize our superclass.
        super().__init__(size=FIXED_LIST_SIZE_MEDIUM)

        # Initialize this type-checking metadata queue.
        self.reinit()


    def reinit(
        self,

        # Optional parameters. Note that these parameters are optional *ONLY* to
        # allow the __init__() method to be trivially defined. In all *OTHER*
        # calls to this method, these parameters should always be passed. Ugh!
        cls_stack: TypeStack = None,
        conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    ) -> None:
        '''
        Reinitialize this type-checking metadata queue.

        Parameters
        ----------
        See the class docstring for further details on passed parameters.
        '''

        # Classify all passed parameters.
        self.conf = conf
        self.cls_stack = cls_stack

        # 1-based indentation level describing the initial level of indentation
        # appropriate for the root hint.
        self.indent_level_child = 1

        # 0-based index of metadata describing the last visitable hint in this
        # queue, initialized to "-1" to ensure that the initial incrementation
        # of this index by the enqueue_hint_child() method initializes index 0
        # of this queue.
        self.index_last = -1

        # Nullify all remaining passed parameters.
        self.exception_prefix = EXCEPTION_PREFIX
        self.func_curr_code = None  # type: ignore[assignment]
        self.func_wrapper_scope = {}
        self.hint_curr_expr = None
        self.hint_curr_meta = None  # type: ignore[assignment]
        self.indent_curr = None  # type: ignore[assignment]
        self.indent_child = None  # type: ignore[assignment]
        self.is_var_random_int_needed = False
        self.pith_curr_expr = None  # type: ignore[assignment]
        self.pith_curr_assign_expr = None  # type: ignore[assignment]
        self.pith_curr_var_name = None  # type: ignore[assignment]
        self.pith_curr_var_name_index = 0

    # ..................{ DUNDERS                            }..................
    def __getitem__(self, hint_index: int) -> HintMeta:  # type: ignore[override]
        '''
        **Type hint type-checking metadata** (i.e., :class:`.HintMeta` object)
        describing the currently visited type hint at the passed index by the
        breadth-first search (BFS) in the
        :func:`beartype._check.code.codemake.make_check_expr` factory.

        For both efficiency and simplicity, this dunder method *always* returns
        a valid :class:`HintMeta` object for all valid indices. This list thus
        behaves similarly to the :class:`collections.defaultdict` container.
        Specifically:

        * If this is the first attempt to access metadata at this index from
          this list (i.e., the value of the item at this index is :data:`None`),
          this dunder method (in order):

          #. Instantiates a new :class:`.HintMeta` object with all fields
             initialized to sane values appropriate for this index.
          #. Replaces the value of the item at this index (which was previously
             :data:`None`) with this new :class:`.HintMeta` object.
          #. Returns new :class:`.HintMeta` object.

        * Else, this is a subsequent access of metadata at this index from this
          list (i.e., the value of the item at this index is an existing
          :class:`.HintMeta` object). In this case, this dunder method simply
          returns that existing :class:`.HintMeta` object as is.

        Parameters
        ----------
        hint_index : int
            0-based absolute index of the type hint type-checking metadata to be
            retrieved, where:

            * Index 0 yields the root type hint currently visited by that BFS.
            * Index 1 yields the first child type hint of that root type hint.
            * And so on.

        Returns
        -------
        HintMeta
            Type hint type-checking metadata at this index.
        '''
        assert isinstance(hint_index, int), f'{repr(hint_index)} not integer.'
        assert 0 <= hint_index <= len(self), (
            f'{hint_index} not in [0, {len(self)}].')

        # Type hint type-checking metadata at this hint_index.
        hint_curr_meta = super().__getitem__(hint_index)  # type: ignore[call-overload]

        # If this metadata has yet to be instantiated...
        if hint_curr_meta is None:
            # Instantiate a new "HintMeta" object with all fields initialized
            # to sane values appropriate for this index.
            hint_curr_meta = self[hint_index] = HintMeta(hint_index=hint_index)
        # Else, this metadata has already been instantiated.

        # Return this metadata.
        return hint_curr_meta

    # ..................{ SETTERS                            }..................
    def set_hint_curr_meta(self, hint_curr_meta: HintMeta) -> None:
        '''
        Set the hint encapsulated by the passed metadata as the currently
        visited hint of the breadth-first search (BFS) iterated by this queue.

        This setter updates instance variables of this queue to reflect that
        this hint is now the currently visited hint.

        Parameters
        ----------
        hint_curr_meta: HintMeta
            Metadata describing the currently visited hint, appended by the
            previously visited parent hint to this queue.
        '''
        assert isinstance(hint_curr_meta, HintMeta), (
            f'{repr(hint_curr_meta)} not "HintMeta" object.')

        # Current level of indentation appropriate for this hint.
        indent_level_curr = hint_curr_meta.indent_level

        # Update instance variables of this queue to reflect that this hint is
        # now the currently visited hint.
        self.hint_curr_meta = hint_curr_meta
        self.indent_level_child = indent_level_curr + 1
        self.indent_curr  = INDENT_LEVEL_TO_CODE[indent_level_curr]
        self.indent_child = INDENT_LEVEL_TO_CODE[self.indent_level_child]
        self.pith_curr_expr = hint_curr_meta.pith_expr
        self.pith_curr_var_name_index = hint_curr_meta.pith_var_name_index
        self.pith_curr_var_name = PITH_INDEX_TO_VAR_NAME[
            self.pith_curr_var_name_index]

        #FIXME: Comment this sanity check out after we're sufficiently
        #convinced this algorithm behaves as expected. While useful, this check
        #requires a linear search over the entire code and is thus costly.
        # assert hint_curr_placeholder in func_wrapper_code, (
        #     '{} {!r} placeholder {} not found in wrapper body:\n{}'.format(
        #         hint_curr_exception_prefix, hint, hint_curr_placeholder, func_wrapper_code))

        # Code snippet type-checking the current pith against this hint.
        self.func_curr_code = None  # type: ignore[assignment]

        # Code expression evaluating to the origin type underlying this hint.
        self.hint_curr_expr = None

    # ..................{ ADDERS                             }..................
    def add_func_scope_type_or_types(
        self, type_or_types: TypeOrSetOrTupleTypes) -> str:
        '''
        Add a new **scoped class or tuple of classes** (i.e., new key-value pair
        of the passed dictionary mapping from the name to value of each globally
        or locally scoped attribute externally accessed elsewhere, whose key is
        a machine-readable name internally generated by this function to
        uniquely refer to the passed class or tuple of classes and whose value
        is that class or tuple) to this local scope of this wrapper function
        *and* return that name.

        This method is merely a high-level convenience wrapping the lower-level
        :func:`beartype._check.code.codescope.add_func_scope_type_or_types`
        function.

        Parameters
        ----------
        type_or_types : TypeOrSetOrTupleTypes
            Classes to be added to this scope, defined as either:

            * A single class.
            * A set of one or more classes.
            * A tuple of one or more classes.

        Returns
        -------
        str
            Name of this class or tuple in this scope generated by this function.

        See Also
        ------
        :func:`beartype._check.code.codescope.add_func_scope_type_or_types`
            Further details.
        '''

        # Defer to the lower-level add_func_scope_type_or_types() adder.
        return add_func_scope_type_or_types(
            type_or_types=type_or_types,
            func_scope=self.func_wrapper_scope,
            exception_prefix=EXCEPTION_PREFIX_FUNC_WRAPPER_LOCAL,
        )

    # ..................{ ENQUEUERS                          }..................
    def enqueue_hint_or_sane_child(
        self, hint_or_sane: HintOrSanifiedData, pith_expr: str) -> str:
        '''
        **Enqueue** (i.e., append) to the end of the this queue new
        **type-checking metadata** (i.e., a :class:`.HintMeta` object)
        describing the currently iterated child type hint with the passed
        metadata, enabling this hint to be visited by the ongoing breadth-first
        search (BFS) traversing over this queue.

        Callers are expected to modify this metadata by modifying these instance
        variables of this higher-level parent object:

        * :attr:`indent_level_child`, the 1-based indentation level describing
          the current level of indentation appropriate for this child hint.
        * :attr:`pith_curr_var_name_index`, the integer suffixing the name of
          each local variable assigned the value of the current pith in a
          assignment expression, thus uniquifying this variable in the body of
          the current wrapper function.

        Parameters
        ----------
        hint_or_sane : HintOrSanifiedData
            Either a type hint *or* **sanified type hint metadata** (i.e.,
            :data:`.HintSanifiedData` object) to be type-checked.
        pith_expr : str
            **Pith expression** (i.e., Python code snippet evaluating to the
            value of) the current **pith** (i.e., possibly nested object of the
            passed parameter or return to be type-checked against the currently
            visited type hint).

        Returns
        -------
        str
            Placeholder string to be subsequently replaced by code type-checking
            this child pith against this child type hint.
        '''

        # Child hint and type variable lookup table encapsulated by this data.
        hint, typevar_to_hint = unpack_hint_or_sane(hint_or_sane)

        #FIXME: This is trash, obviously. Instead, this should probably be
        #returned by the unpack_hint_or_sane() function.
        # Recursion guard (i.e., frozen set of the integers uniquely identifying
        # *ALL* transitive recursable parent hints of this hint), defined as
        # either...
        recursable_hint_ids: FrozenSetInts = FROZENSET_EMPTY
        #     # If there is *NO* currently visited hint, then the passed hint is
        #     # the root hint and thus has *NO* parent hint. In this case,
        #     # initialize this recursion guard to the empty frozen set. The first
        #     # iteration of the parent make_check_expr() code factory calling
        #     # this method will then ensure that the follownig "else" branch will
        #     # produce the first non-empty recursion guard resembling:
        #     #     FROZENSET_EMPTY | {id(root_hint)} ==
        #     #     frozenset((id(root_hint),))
        #     FROZENSET_EMPTY
        #     if self.hint_curr_meta is None else
        #     # Else, a parent hint of this child hint is currently being visited.
        #     # In this case, produce the frozen set of the integers uniquely
        #     # identifying *ALL* transitive parent hints of this child hint
        #     # (including this parent hint of this child hint) by:
        #     # * Efficiently extending the frozen set of the integers uniquely
        #     #   identifying *ALL* transitive parent hints of this parent hint by
        #     #   the integer uniquely identifying this parent hint.
        #     #
        #     # Note that OR-ing a "frozenset" with a "set" produces yet another
        #     # "frozenset" and is, indeed, the most efficient means of doing so:
        #     #     >>> frozenset(('ok',)) | {'ko',}
        #     #     frozenset({'ok', 'ko'})
        #     self.hint_curr_meta.parent_hint_ids | {id(
        #         self.hint_curr_meta.hint),}
        # )

        # Return the placeholder string to be subsequently replaced by code
        # type-checking this child pith against this child hint, produced by
        # enqueueing new type-checking metadata describing this child hint.
        return self._enqueue_hint_child(
            hint=hint,
            indent_level=self.indent_level_child,
            pith_expr=pith_expr,
            pith_var_name_index=self.pith_curr_var_name_index,
            recursable_hint_ids=recursable_hint_ids,
            typevar_to_hint=typevar_to_hint,
        )


    #FIXME: Possibly just inline into the above method call. Previously, we
    #required these to be separate methods. Now we no longer do. We sigh. *sigh*
    def _enqueue_hint_child(self, **kwargs) -> str:
        '''
        **Enqueue** (i.e., append) to the end of the this queue new
        **type-checking metadata** (i.e., a :class:`.HintMeta` object)
        describing the currently iterated child type hint with the passed
        metadata, enabling this hint to be visited by the ongoing breadth-first
        search (BFS) traversing over this queue.

        Parameters
        ----------
        All passed keyword parameters are passed as is to the lower-level
        :meth:`.HintMeta.reinit` method.

        Returns
        -------
        str
            Placeholder string to be subsequently replaced by code type-checking
            this child pith against this child type hint.
        '''
        # print(f'Enqueing child hint {self.index_last+1} with {repr(kwargs)}...')

        # Increment the 0-based index of metadata describing the last visitable
        # hint in this list (which also serves as the unique identifier of the
        # currently iterated child hint) *BEFORE* overwriting the existing
        # metadata at this index.
        #
        # Note this index is guaranteed to *NOT* exceed the fixed length of this
        # list. By prior validation, "FIXED_LIST_SIZE_MEDIUM" is guaranteed to
        # be substantially larger than "hints_meta_index_last".
        self.index_last += 1

        # Type hint type-checking metadata at this index.
        hint_meta = self.__getitem__(self.index_last)

        # Replace prior fields of this metadata with the passed fields.
        hint_meta.reinit(**kwargs)

        # Return the placeholder substring associated with this type hint.
        return hint_meta.hint_placeholder

    # ..................{ SANIFIERS                          }..................
    def sanify_hint_child(
        self,

        # Mandatory parameters.
        hint: Hint,

        # Optional parameters.
        typevar_to_hint: Optional[TypeVarToHint] = None,
    ) -> HintOrSanifiedData:
        '''
        Type hint sanified (i.e., sanitized) from the passed **possibly insane
        child type hint** (i.e., possibly PEP-noncompliant hint transitively
        subscripting the root type hint annotating a parameter or return of the
        currently decorated callable) if this hint is both reducible and
        unignorable, this hint unmodified if this hint is both irreducible and
        unignorable, or :obj:`typing.Any` otherwise (i.e., if this hint is
        ignorable).

        This method is merely a convenience wrapper for the lower-level
        :func:`.sanify_hint_child` sanifier.

        Parameters
        ----------
        hint : Hint
            Child type hint to be sanified.
        typevar_to_hint : TypeVarToHint, optional
            **Type variable lookup table** (i.e., immutable dictionary mapping
            from the :pep:`484`-compliant **type variables** (i.e.,
            :class:`typing.TypeVar` objects) originally parametrizing the
            origins of all transitive parent hints of this hint to the
            corresponding child hints subscripting these parent hints). Defaults
            to :data:`None`, in which case this table actually defaults to that
            of the currently visited parent hint of this child hint (i.e.,
            ``self.hint_curr_meta.typevar_to_hint``).

        Returns
        -------
        HintOrSanifiedData
            Either:

            * If this child hint is reducible to:

              * An ignorable lower-level hint, :obj:`typing.Any`.
              * An unignorable lower-level hint, either:

                * If reducing this hint to that lower-level hint generates
                  supplementary metadata, that metadata.
                * Else, that lower-level hint alone.

            * Else, this child hint is irreducible. In this case, this child
              hint unmodified.
        '''
        assert isinstance(typevar_to_hint, NoneTypeOr[FrozenDict]), (
            f'{repr(typevar_to_hint)} neither frozen dictionary nor "None".')

        #FIXME: *NON-IDEAL.* Ideally, @beartype would actually generate code
        #recursively type-checking recursive hints. However, doing so is
        #*EXTREMELY* non-trivial. Why?
        #
        #Non-triviality is one obvious concern. For each recursive hint,
        #@beartype must now:
        #* Dynamically generate one low-level recursive type-checking function
        #  unique to that recursive hint.
        #* Call each such function in higher-level wrapper functions to
        #  type-check each pith against the corresponding recursive hint.
        #
        #Safety is another obvious concern. Generated code *MUST* explicitly
        #guard against infinitely recursive containers:
        #    >>> infinite_list = []
        #    >>> infinite_list.append(infinite_list)  # <-- gg fam
        #
        #But guarding against infinitely recursive containers requires
        #maintaining a (...waitforit) frozen set of the IDs of all previously
        #type-checked objects, which must then be passed to each dynamically
        #generated recursive type-checking function that type-checks a specific
        #recursive hint. Maintaining these frozen sets then incurs a probably
        #significant space and time complexity hit.
        #
        #In short, it's pretty brutal stuff. For now, simply ignoring recursion
        #strikes us the sanest and certainly simplest approach. *sigh*

        #FIXME: Clean up *ALL* references to "parent_hint_ids", please. *sigh*
        #FIXME: Unit test us up, please.
        # If the integer identifying this child hint is that of a transitive
        # parent hint of this child hint, this child hint has already been
        # visited by the current breadth-first search (BFS) and thus constitutes
        # a recursive hint. Certainly, various approaches to generating code
        # type-checking recursive hints exists. @beartype currently embraces the
        # easiest, fastest, and laziest approach: simply ignore all recursion!

        #FIXME: *HMMMM.* Right. This needs to be tested *AFTER* sanification,
        #really. Either that, or we need to be adding pre-sanified hint IDs. The
        #point is that we need to be consistent about what we're adding and
        #testing. Post-sanified IDs is probably best, honestly.
        #FIXME: Actually, this should be tested in the existing
        #reduce_hint_pep695_unsubscripted() reducer *BEFORE* sanification,
        #obviously.

        # if id(hint) in self.hint_curr_meta.parent_hint_ids:
        #     return ANY
        # Else, this child hint has *NOT* yet been visited by this BFS.

        # If *NO* type variable lookup table was passed, default this table to
        # that of of the currently visited parent hint of this child hint.
        if typevar_to_hint is None:
            typevar_to_hint = self.hint_curr_meta.typevar_to_hint
        # Else, a type variable lookup table was passed. In this case, preserve
        # this table as is.

        # Sane hint sanified from this possibly insane hint if sanifying this
        # hint did not generate supplementary metadata *OR* that metadata
        # otherwise (i.e., if doing so generated supplementary metadata).
        hint_or_sane_child = sanify_hint_child(
            hint=hint,
            cls_stack=self.cls_stack,
            conf=self.conf,
            typevar_to_hint=typevar_to_hint,
            exception_prefix=self.exception_prefix,
        )

        # Return this metadata.
        return hint_or_sane_child



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`- or :pep:`604`-compliant **union type-checking code
factories** (i.e., low-level callables dynamically generating pure-Python code
snippets type-checking arbitrary objects against union type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Validate that PEP 695-compliant type aliases aliasing recursive unions
#behave as expected: e.g.,
#    type recursive_union = int | recursive_union
#FIXME: Likewise, note that our current *EXTREMELY* non-trivial handling of
#"hint_overrides"-based recursive unions below could probably benefit from being
#refactored into the same approach used to handle PEP 695-based recursive
#unions. The approach below is wild -- and not the good kind of "wild," either.

# ....................{ IMPORTS                            }....................
from beartype.typing import Tuple
from beartype._check.metadata.hint.hintsmeta import HintsMeta
from beartype._check.metadata.metasane import (
    HintOrSanifiedData,
    HintSanifiedData,
    DictHintOrSanifiedDataToAny,
    ListHintOrSanifiedData,
    SetHintOrSanifiedData,
    get_hint_or_sane_hint,
    unpack_hint_or_sane,
)
from beartype._data.code.datacodemagic import LINE_RSTRIP_INDEX_OR
from beartype._data.code.pep.datacodepep484604 import (
    CODE_PEP484604_UNION_CHILD_PEP_format,
    CODE_PEP484604_UNION_CHILD_NONPEP_format,
    CODE_PEP484604_UNION_PREFIX,
    CODE_PEP484604_UNION_SUFFIX,
)
# from beartype._data.hint.datahintpep import SetHints
from beartype._data.hint.datahinttyping import DictTypeToAny
from beartype._data.hint.pep.sign.datapepsignset import HINT_SIGNS_UNION
from beartype._util.cache.pool.utilcachepoolinstance import (
    acquire_instance,
    release_instance,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.utilpepget import (
    get_hint_pep_args,
    get_hint_pep_sign_or_none,
)
from beartype._util.hint.pep.utilpeptest import is_hint_pep

# ....................{ FACTORIES                          }....................
def make_hint_pep484604_check_expr(hints_meta: HintsMeta) -> None:
    '''
    Either a Python code snippet type-checking the current pith against the
    passed :pep:`484`- or :pep:`604`-compliant union type hint if this union is
    **unignorable** (i.e., subscripted by *no* ignorable child type hints) *or*
    :data:`None` otherwise (i.e., if this union is ignorable).

    This factory is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), due to accepting **context-sensitive
    parameters** (i.e., whose values contextually depend on context unique to
    the code being generated for the currently decorated callable) such as
    ``pith_curr_assign_expr``.

    Caveats
    -------
    Unions are non-physical abstractions of physical types and thus *not*
    themselves subject to type-checking; only the subscripted arguments of
    unions are type-checked. This differs from :mod:`typing` pseudo-containers
    like ``List[int]``, in which both the parent :obj:`typing.List` and child
    :class:`int` types represent physical types to be type-checked. Ergo, unions
    themselves impose no narrowing of the current pith expression and thus
    *cannot* by definition benefit from assignment expressions. This differs
    from :mod:`typing` pseudo-containers, which narrow the current pith
    expression and thus do benefit from assignment expressions.

    Parameters
    ----------
    hint_meta : HintMeta
        Metadata describing the currently visited hint, appended by the
        previously visited parent hint to the ``hints_meta`` stack.
    hints_meta : HintsMeta
        Stack of metadata describing all visitable hints currently discovered by
        this breadth-first search (BFS).
    cls_stack : TypeStack, optional
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    func_wrapper_scope : LexicalScope
        Local scope (i.e., dictionary mapping from the name to value of each
        attribute referenced in the signature) of this wrapper function required
        by this Python code snippet.
    pith_curr_expr : str
        Full Python expression evaluating to the value of the **current pith**
        (i.e., possibly nested object of the current parameter or return value
        to be type-checked against this union type hint).
    pith_curr_assign_expr : str
        Assignment expression assigning this full Python expression to the
        unique local variable assigned the value of this expression.
    pith_curr_var_name_index : int
        Integer suffixing the name of each local variable assigned the value of
        the current pith in a assignment expression, thus uniquifying this
        variable in the body of the current wrapper function.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.
    '''
    assert isinstance(hints_meta, HintsMeta), (
        f'{repr(hints_meta)} not "HintsMeta" object.')

    # ....................{ LOCALS                         }....................
    # Flattened tuple of two or more child hints subscripting this parent union
    # such that *ALL* nested child hints subscripting child unions are expanded
    # directly into this tuple, thus non-destructively eliminating child unions.
    #
    # Note that this getter is memoized and thus requires:
    # * Positional parameters.
    # * "hint_meta" instance variables to be explicitly passed rather than the
    #   "hint_meta" object in entirety. Why? Memoization, of course. Passing the
    #   "hint_meta" object in entirety would effectively inhibit the memoization
    #   of this getter, which entirely defeats the point.
    hint_or_sane_childs = _get_hint_pep484604_union_args_flattened(hints_meta)
    # print(f'Unflattened union {repr(hints_meta.hint_curr_meta.hint)}...')
    # print(f'...child hints {repr(hint_or_sane_childs)}.')

    # Dictionary whose keys comprise the set of all PEP-noncompliant child hints
    # subscripting this union and whose values are ignorable. Whereas sets fail
    # to preserve insertion order, dictionaries preserve insertion order. While
    # a set would be preferable for simplicity, the fact that sets are currently
    # unordered leaves us no choice but to abuse dictionaries.
    #
    # The order of union child hints is insignificant in the common case but
    # *NOT* the worst case, where order is significant. Examples include:
    # * A union over classes whose metaclasses define custom type-checking by
    #   overriding the __instancecheck__() dunder method. Callers might
    #   intentionally order those classes in that union to maximize
    #   type-checking efficiency. How? In descending order of the efficiency of
    #   those __instancecheck__() implementations. The most efficiently
    #   type-checkable classes would be listed first in that union; the least
    #   efficiently type-checkable classes, last.
    # * A union over classes registered as pseudo-subclasses with the
    #   collections.abc.ABCMeta.register() method. Callers might intentionally
    #   order those classes in that union to prevent type-checking code
    #   dynamically generated by beartype from raising unexpected exceptions in
    #   the event that those classes fail to fully satisfy the API of the ABC
    #   they were registered with. See also this issue:
    #       https://github.com/beartype/beartype/issues/499#issuecomment-2683400721
    #
    # Since PEP-compliant and -noncompliant child hints require fundamentally
    # different forms of type-checking, prefiltering child hints into these
    # dictionaries *BEFORE* generating code type-checking these child hints
    # improves both efficiency and maintainability.
    hint_childs_nonpep: DictTypeToAny = acquire_instance(dict)

    # Dictionary whose keys comprise the set of all PEP-compliant child hints
    # subscripting this union and whose values are ignorable. See above.
    hint_or_sane_childs_pep: DictHintOrSanifiedDataToAny = acquire_instance(
        dict)

    # ....................{ FILTER                         }....................
    #FIXME: Optimize by refactoring into a "while" loop. Naturally, profile that
    #doing so actually *IS* an optimization before doing so. *sigh*

    # For each child hint subscripting this union...
    for hint_or_sane_child in hint_or_sane_childs:
        #FIXME: Uncomment as desired for debugging. This test is currently a bit
        #too costly to warrant uncommenting.
        # Assert that this child hint is *NOT* shallowly ignorable. Why? Because
        # any union containing one or more shallowly ignorable child hints is
        # deeply ignorable and should thus have already been ignored after a
        # call to the is_hint_ignorable() tester passed this union on handling
        # the parent hint of this union.
        # assert (
        #     repr(hint_curr) not in HINTS_REPR_IGNORABLE_SHALLOW), (
        #     f'{hint_curr_exception_prefix} {repr(hint_curr)} child '
        #     f'{repr(hint_child)} ignorable but not ignored.')

        # Child hint encapsulated by this metadata.
        hint_child = get_hint_or_sane_hint(hint_or_sane_child)

        # If this child hint is PEP-compliant, filter this child hint *AND* all
        # associated metadata (if any) into this dictionary of PEP-compliant
        # child hints.
        #
        # Note that this PEP-compliant child hint *CANNOT* also be filtered into
        # the dictionary of PEP-noncompliant child hints, even if this child
        # hint originates from a non-"typing" type (e.g., "List[int]" from
        # "list"). Why? Because that would then induce false positives when the
        # current pith shallowly satisfies this non-"typing" type but does *NOT*
        # deeply satisfy this child hint.
        if is_hint_pep(hint_child):
            hint_or_sane_childs_pep[hint_or_sane_child] = None
        # Else, this child hint is PEP-noncompliant. In this case, filter this
        # child hint into this dictionary of PEP-noncompliant child hints. Since
        # PEP-noncompliant hints are by definition associated with *NO*
        # meaningful metadata, silently ignore this metadata.
        else:
            hint_childs_nonpep[hint_child] = None  # type: ignore[index]

    # ....................{ NON-PEP                        }....................
    # Initialize the code type-checking the current pith against these arguments
    # to the substring prefixing all such code.
    hints_meta.func_curr_code = CODE_PEP484604_UNION_PREFIX

    # If this union is subscripted by one or more PEP-noncompliant child hints,
    # generate and append efficient code type-checking these child hints
    # *BEFORE* less efficient code type-checking any PEP-compliant child hints
    # subscripting this union.
    if hint_childs_nonpep:
        hints_meta.func_curr_code += (
            CODE_PEP484604_UNION_CHILD_NONPEP_format(
                # Python expression yielding the value of the current pith.
                # Specifically...
                pith_curr_expr=(
                    # If this union is also subscripted by one or more
                    # PEP-compliant child hints, prefer the expression assigning
                    # this value to a local variable efficiently reused by
                    # subsequent code generated for those PEP-compliant child
                    # hints.
                    hints_meta.pith_curr_assign_expr
                    if hint_or_sane_childs_pep else
                    # Else, this union is subscripted by *NO* PEP-compliant
                    # child hints. Since this is the first and only test
                    # generated for this union, prefer the expression yielding
                    # the value of the current pith *WITHOUT* assigning this
                    # value to a local variable, which would needlessly go
                    # unused.
                    hints_meta.pith_curr_expr
                ),
                # Python expression evaluating to a tuple of these arguments.
                #
                # Note that we would ideally avoid coercing this set into a
                # tuple when this set only contains one type by passing that
                # type directly to the _add_func_wrapper_local_type() function.
                # Sadly, the "set" class defines no convenient or efficient
                # means of retrieving the only item of a 1-set. Indeed, the most
                # efficient means of doing so is to iterate over that set and
                # immediately halt iteration:
                #     for first_item in muh_set: break
                #
                # While we *COULD* technically leverage that approach here,
                # doing so would also mandate adding multiple intermediate
                # tests, mitigating any performance gains. Ultimately, we avoid
                # doing so by falling back to the usual approach. See also this
                # relevant self-StackOverflow post:
                #       https://stackoverflow.com/a/40054478/2809027
                hint_curr_expr=hints_meta.add_func_scope_type_or_types(
                    hint_childs_nonpep.keys()),
            ))

    # ....................{ PEP                            }....................
    # For the 0-based index of each PEP-compliant child hint of this union *AND*
    # that hint...
    for hint_or_sane_child_pep_index, hint_or_sane_child_pep in enumerate(
        hint_or_sane_childs_pep.keys()):
        # print(f'Enqueing union {repr(hints_meta.hint_curr_meta.hint)}...')
        # print(f'...PEP-compliant child {repr(hint_or_sane_child_pep)}.')

        # Code deeply type-checking this child hint.
        hints_meta.func_curr_code += CODE_PEP484604_UNION_CHILD_PEP_format(
            # Expression yielding the value of this pith.
            hint_child_placeholder=hints_meta.enqueue_hint_or_sane_child(
                hint_or_sane=hint_or_sane_child_pep,
                pith_expr=(
                    # If either...
                    #
                    # Then prefer the expression efficiently reusing the value
                    # previously assigned to a local variable by either the
                    # above conditional or prior iteration of the current
                    # conditional.
                    hints_meta.pith_curr_var_name
                    if (
                        # This union is also subscripted by one or more
                        # PEP-noncompliant child hints *OR*...
                        hint_childs_nonpep or
                        # This is any PEP-compliant child hint *EXCEPT* the
                        # first...
                        hint_or_sane_child_pep_index
                    ) else
                    # Then this union is not subscripted by any PEP-noncompliant
                    # child hints *AND* this is the first PEP-compliant child
                    # hint. In this case, preface this code with an expression
                    # assigning this value to a local variable efficiently
                    # reused by code generated by subsequent iteration.
                    #
                    # Note this child hint is guaranteed to be followed by at
                    # least one more child hint. Why? Because the "typing"
                    # module forces unions to be subscripted by two or more
                    # child hints. By deduction, those child hints *MUST* be
                    # PEP-compliant. Ergo, we need *NOT* explicitly validate
                    # that constraint here.
                    hints_meta.pith_curr_assign_expr
                ),
            ),
        )

    # ....................{ RETURN                         }....................
    # Release this pair of sets back to their respective pools.
    release_instance(hint_childs_nonpep)
    release_instance(hint_or_sane_childs_pep)

    # If this code is *NOT* its initial value, this union is subscripted by one
    # or more unignorable child hints and the above logic generated code
    # type-checking these child hints. In this case...
    if hints_meta.func_curr_code is not CODE_PEP484604_UNION_PREFIX:
        # Munge this code to...
        hints_meta.func_curr_code = (
            # Strip the erroneous " or" suffix appended by the last child hint
            # from this code.
            f'{hints_meta.func_curr_code[:LINE_RSTRIP_INDEX_OR]}'
            # Suffix this code by the substring suffixing all such code.
            f'{CODE_PEP484604_UNION_SUFFIX}'
        # Format the "indent_curr" prefix into this code, deferred above for
        # efficiency.
        ).format(indent_curr=hints_meta.indent_curr)
    # Else, this snippet is its initial value and thus ignorable.

# ....................{ PRIVATE ~ getters                  }....................
@callable_cached
def _get_hint_pep484604_union_args_flattened(
    hints_meta: HintsMeta) -> Tuple[HintOrSanifiedData, ...]:
    '''
    Flattened tuple of the two or more child hints subscripting the passed
    :pep:`604`- or :pep:`484`-compliant union hint such that *all* nested child
    hints subscripting *all* child union hints are **flattened** (i.e., expanded
    directly) into the returned tuple, thus non-destructively eliminating *all*
    child union hints of this parent union hint.

    This getter recursively flattens arbitrarily nested child union hints
    regardless of nesting depth in this parent union hint. Although non-trivial,
    doing so is required to support uncommon edge cases -- including:

    * :`pep:`695`-compliant type aliases, which technically enable users to
      deeply nest type aliases and thus unions to an arbitrarily deep nesting
      level: e.g.,

      .. code-block:: python

         # After reduction, the PEP 695-compliant "Level3" type alias defined
         # below effectively resembles:
         #     type Level3 = ((int | complex) | str) | float
         type Level1 = int | complex
         type Level2 = Level1 | str
         type Level3 = Level2 | float

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the only function calling this
    getter *is* itself memoized.

    Caveats
    -------
    **This getter exhibits time complexity** :math:`O(k)` **for k the total
    number of all transitive child hints of this union hint across all deeply
    nested child unions of this parent union hint.** A looser bound but somewhat
    more intelligible upper complexity bound would be :math:`O(n**m)`, where:

    * :m.th:`n` is the maximum number of child hints in any deeply nested child
      union of this parent union hint.
    * :math:`m` is the maximum depth of the most deeply nested child union of
      this parent union hint.

    Parameters
    ----------
    hints_meta : HintsMeta
        **Type hint type-checking metadata queue** (i.e., low-level fixed list
        of metadata describing all visitable type hints currently discovered by
        the breadth-first search (BFS) dynamically generating pure-Python
        type-checking code snippets in the
        :func:`beartype._check.code.codemake.make_check_expr` factory).

    Returns
    -------
    Tuple[HintOrSanifiedData, ...]
        Flattened tuple of the two or more child hints *or* **sanified child
        hint metadatum** (i.e., :class:`.HintSanifiedData` objects) subscripting
        this parent union hint.

    Raises
    ------
    BeartypeDecorHintPep604Exception
        If this tuple is empty.
    '''
    # print(f'[484/604] hint_curr_meta: {repr(hints_meta.hint_curr_meta)}')

    # ....................{ LOCALS                         }....................
    # Metadata localized from this dataclass for both usability and efficiency.
    hint = hints_meta.hint_curr_meta.hint
    typevar_to_hint = hints_meta.hint_curr_meta.typevar_to_hint

    # ....................{ LOCALS ~ child                 }....................
    # Tuple of the two or more child hints subscripting this union.
    hint_childs = get_hint_pep_args(hint)

    # Number of these child hints.
    hint_childs_len = len(hint_childs)

    # Assert this union to be subscripted by two or more child hints.
    #
    # Note this should *ALWAYS* be the case, as:
    # * The unsubscripted "typing.Union" type hint factory is explicitly listed
    #   in the "HINTS_REPR_IGNORABLE_SHALLOW" set and should thus have already
    #   been ignored when present.
    # * The "typing" module explicitly prohibits empty union subscription: e.g.,
    #       >>> typing.Union[]
    #       SyntaxError: invalid syntax
    #       >>> typing.Union[()]
    #       TypeError: Cannot take a Union of no types.
    # * The "typing" module reduces unions of one child hint to that hint: e.g.,
    #     >>> import typing
    #     >>> typing.Union[int]
    #     int
    assert hint_childs_len >= 2, (
        f'{hints_meta.exception_prefix}'
        f'PEP 484 or 604 union type hint {repr(hint)} either unsubscripted '
        f'or subscripted by only one child type hint.'
    )

    # ....................{ LOCALS ~ list                  }....................
    # List of all currently sanified child hints from which to reconstitute this
    # union, sanified by the current iteration of the loop performed below.

    # Input stack of all currently unflattened transitive child hints of this
    # union to be visited by the depth-first search (DFS) below, initialized to
    # the non-empty list of all direct child hints of this union.
    hint_or_insane_childs_unflattened: ListHintOrSanifiedData = (
        acquire_instance(list))
    hint_or_insane_childs_unflattened.extend(hint_childs)

    # Output stack of all previously flattened transitive child hints of this
    # union that have already been visited by this DFS, initialized to the
    # empty list.
    #
    # Note that this stack orders these child hints in the *REVERSE* order that
    # these child hints were originally ordered by the user in this union.
    hint_or_sane_childs_flattened: ListHintOrSanifiedData = (
        acquire_instance(list))

    # ....................{ LOCALS ~ set                   }....................
    # Set of all previously sanified child hints, required to avoid infinite
    # recursion that could otherwise be induced in worst-case sanification --
    # including:
    # * If the caller overrode at least one type hint to expand to a union
    #   recursively embedding the same type hint through a beartype
    #   configuration enabling the "hint_overrides" option (e.g., a
    #   "hints_meta.conf.hint_overrides" option whose value is
    #   "BeartypeHintOverrides({int: int | float})".
    hint_or_insane_childs_sanified: SetHintOrSanifiedData = (
        acquire_instance(set))

    # ....................{ SEARCH                         }....................
    # Repeatedly flatten all currently unflattened transitive child hints of
    # this union into an increasingly flat union until this union can no longer
    # be flattened (i.e., after these child hints have all been flattened).
    # Recursively perform a depth-first search (DFS) over these child hints with
    # non-recursive iteration for both efficiency and simplicity. While
    # non-recursive iteration is often non-trivial, this implementation is
    # surprisingly trivial.
    #
    # While the input stack of these child hints is still non-empty, one or more
    # transitive child hints of this union have yet to be flattened.
    while hint_or_insane_childs_unflattened:
        # ....................{ SANIFY                     }....................
        # Currently unflattened transitive child hint to be flattened, defined
        # as either:
        # * If either a nested parent union of this child hint has not already
        #   been sanified by a prior iteration of this loop *OR* a nested parent
        #   union of this child hint has already been sanified by a prior
        #   iteration of this loop but doing so did not generate supplementary
        #   metadata, this child hint as is.
        # * Else, sanifying this child hint generated supplementary metadata. In
        #   this case, that metadata.
        hint_or_insane_child = hint_or_insane_childs_unflattened.pop()

        # True only if this child hint has *NOT* already been sanified by a
        # previously performed sanification in this recursive tree of all
        # previously performed sanifications. For safety, assume insanity.
        is_hint_child_insane = True

        # True only if this child hint was sanified via a hint override
        # externally configured by the user. For safety, assume *NO* override.
        is_hint_child_overridden = False

        # Attempt to...
        #
        # Note that the is_object_hashable() tester is internally implemented
        # with the same Easier to Ask for Permission than Forgiveness
        # (EAFP)-based "try-except" block and is thus equally inefficient. In
        # fact, the current approach avoids an extraneous call to that tester
        # and is thus marginally faster. (Emphasis on "marginally.")
        try:
            # True only if this child hint has *NOT* already been sanified by a
            # previously performed sanification in this recursive tree of all
            # previously performed sanifications.
            is_hint_child_insane = (
                hint_or_insane_child not in hint_or_insane_childs_sanified)

            # If this child hint has yet to be sanified...
            if is_hint_child_insane:
                # Record this child hint as now having been sanified *AFTER*
                # detecting whether this child hint requires sanification. Ugh!
                #
                # Note that adding an item to a set requires that item to be
                # hashable. For safety, this addition is intentionally isolated
                # to this "try" branch.
                hint_or_insane_childs_sanified.add(hint_or_insane_child)
            # Else, this child hint has already been sanified. In this case...
            else:
                # Sane child hint encapsulated by this metadata.
                hint_child_sane = get_hint_or_sane_hint(hint_or_insane_child)

                # True only if this child hint was sanified via a hint override
                # externally configured by the end user.
                is_hint_child_overridden = (
                    hint_child_sane in hints_meta.conf.hint_overrides)
        # If doing so raises a "TypeError", this child hint is unhashable. In
        # this case, assume this child hint to be insane. You know who you are.
        except TypeError:
            pass

        # If this child hint has yet to be sanified...
        if is_hint_child_insane:
            # Possibly insane child hint and corresponding type variable lookup
            # table encapsulated by this metadata.
            hint_child, hint_child_typevar_to_hint = unpack_hint_or_sane(
                hint_or_sane=hint_or_insane_child,
                typevar_to_hint=typevar_to_hint,
            )

            # Sane child hint sanified from this possibly insane child hint
            # if sanifying this child hint did not generate supplementary
            # metadata *OR* that metadata otherwise (i.e., if sanifying this
            # child hint generated supplementary metadata).
            #
            # Note that this sanification is intentionally performed *BEFORE*
            # the sign of this child hint is tested. Why? Because some
            # reductions expand an arbitrary hint into a union. This includes:
            # * PEP 695-compliant type aliases aliased to unions. See above!
            # * The PEP-noncompliant "float' and "complex" types, implicitly
            #   expanded to the PEP 484-compliant "float | int" and "complex |
            #   float | int" type hints (respectively) when the non-default
            #   "conf.is_pep484_tower=True" parameter is enabled.
            # * User-defined "BeartypeHintOverrides", a generalization of the
            #   prior item.
            # print(f'Sanifying union child hint {repr(hint_child)} under {repr(conf)}...')
            hint_or_sane_child = hints_meta.sanify_hint_child(
                hint=hint_child, typevar_to_hint=hint_child_typevar_to_hint)
            # print(f'Sanified union child hint to {repr(hint_or_sane_child)}.')
        # Else, this child hint has already been sanified by a previously
        # performed sanification in this recursive tree of all previously
        # performed sanifications.
        #
        # There now exist two divergent edge cases. Either:
        # * This child hint has intrinsic value and *MUST* thus be preserved as
        #   a member of this union. This edge case currently arises *ONLY*
        #   through hint overrides. Consider overriding the child hint "float"
        #   with the PEP 604-compliant union "int | float". In this case:
        #   * The initial sanification of the child hint "float" performed above
        #     will expand that hint to "int | float".
        #   * The unflattening performed below will then append the child hints
        #     "int" and "float" for subsequent iteration by this "while" loop.
        #   * The child hint "float" will then be detected as having been
        #     previously sanified. This child hint *CANNOT* be removed or
        #     ignored, as doing so would omit this child hint from this union.
        #     This child hint thus has intrinsic value and *MUST* thus be
        #     preserved as a member of this union.
        # * This child hint lacks intrinsic value and *MUST* thus be removed as
        #   a member of this union. This edge case currently arises *ONLY*
        #   through PEP 695-compliant recursive type aliases. Consider:
        #       type RecursiveAlias = int | RecursiveAlias
        #
        #   The recursive type alias "RecursiveAlias" merely has symbolic value
        #   and thus lacks intrinsic value. This type alias is semantically
        #   equivalent to the non-recursive type alias:
        #       type NonrecursiveAlias = int
        #
        #   Similar logic as with the prior edge case then applies, except that
        #   this hild *SHOULD* be removed and ignored (rather than preserved).
        #
        # If this child hint was overridden by an external hint override, this
        # child hint has itrinsic value and *MUST* thus be preserved. Avoid
        # sanifying this child hint yet again, as doing so would provoke
        # infinite recursion. Reuse this previously sanified child hint as is.
        elif is_hint_child_overridden:
            hint_or_sane_child = hint_or_insane_child
        # Else, this child hint was *NOT* overridden by an external hint
        # override. In this case, this child hint lacks intrinsic value.
        # Fortunately, this child hint has already been sanified and thus
        # appended to this union as a member. In this case, ignoring this
        # extraneous child hint here is both safe and desirable.
        else:
            continue

        # ....................{ UNION                      }....................
        # Sane child hint encapsulated by this metadata.
        hint_child = get_hint_or_sane_hint(hint_or_sane_child)

        # Sign of this sanified child hint if this hint is PEP-compliant *OR*
        # "None" otherwise (i.e., if this hint is PEP-noncompliant).
        hint_child_sign = get_hint_pep_sign_or_none(hint_child)

        # If this child hint is itself a child union nested in this parent
        # union, explicitly flatten this nested union by appending *ALL* child
        # child hints subscripting this child union onto this parent union.
        #
        # Note that this edge case currently *ONLY* arises when this child hint
        # has been expanded by the above call to the sanify_hint_child()
        # function from a non-union (e.g., "float") into a union (e.g., "float |
        # int"). The standard PEP 484-compliant "typing.Union" type hint factory
        # already implicitly flattens nested unions: e.g.,
        #     >>> from typing import Union
        #     >>> Union[float, Union[int, str]]
        #     typing.Union[float, int, str]
        if hint_child_sign in HINT_SIGNS_UNION:
            # Tuple of all child child hints subscripting this child union.
            hint_child_childs = get_hint_pep_args(hint_child)
            # print(f'Expanding union {repr(hint_curr)} with child union {repr(hint_child_childs)}...')

            # If this child union is encapsulated by metadata...
            if isinstance(hint_or_sane_child, HintSanifiedData):
                # For each child child type subscripting this child union...
                for hint_child_child in hint_child_childs:
                    # Metadata encapsulating the sanification of this child
                    # child hint.
                    hint_or_sane_child_child = hint_or_sane_child.permute(
                        hint=hint_child_child)

                    # Inefficiently append this child child hint to this parent
                    # union in a manner preserving this metadata.
                    hint_or_insane_childs_unflattened.append(
                        hint_or_sane_child_child)
            # Else, this is a bare hint *NOT* encapsulated by metadata. In this
            # case...
            else:
                # Efficiently append these child child hints to this parent
                # union while ignoring this non-existent metadata.
                hint_or_insane_childs_unflattened.extend(hint_child_childs)
        # ....................{ NON-UNION                  }....................
        # Else, this child hint is *NOT* itself a union. In this case, append
        # this child hint to this parent union.
        else:
            hint_or_sane_childs_flattened.append(hint_or_sane_child)
            # print(f'Flattened union child {repr(hint_child)} to {repr(hint_or_sane_child)}...')

    # ....................{ RETURN                         }....................
    # Tuple of all flattened transitive child hints of this union, reversed so
    # as to undo the reverse ordering applied by the above algorithm and thus
    # preserve the original ordering of these child hints of this union.
    hint_or_sane_childs_flattened_tuple = tuple(reversed(
        hint_or_sane_childs_flattened))

    # Release these stacks back to their respective pool.
    release_instance(hint_or_insane_childs_sanified)
    release_instance(hint_or_insane_childs_unflattened)
    release_instance(hint_or_sane_childs_flattened)
    # print(f'Flattened union to {repr(hint_or_sane_childs)}...')

    # Return this tuple.
    return hint_or_sane_childs_flattened_tuple



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype sanified type hint metadata dataclass** (i.e., class aggregating
*all* metadata returned by :mod:`beartype._check.convert.convsanify` functions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeDecorHintSanifyException
from beartype.typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Set,
    Tuple,
    Union,
)
from beartype._data.hint.datahintpep import (
    Hint,
    TypeVarToHint,
)
from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._util.kind.map.utilmapfrozen import FrozenDict
from beartype._util.utilobjmake import permute_object

# ....................{ CLASSES                            }....................
#FIXME: Unit test us up, please.
class HintSanifiedData(object):
    '''
    **Beartype sanified type hint metadata** (i.e., immutable and thus hashable
    object encapsulating *all* metadata returned by
    :mod:`beartype._check.convert.convsanify` functions after sanitizing a
    possibly PEP-noncompliant type hint into a fully PEP-compliant type hint).

    For efficiency, these functions only conditionally return this metadata for
    the proper subset of type hints that actually modify this metadata; since
    most type hints do *not* modify this metadata, these functions typically
    only return the sanified type hint (rather than both that hint and all of
    this other ancillary metadata).

    Attributes
    ----------
    hint : Hint
        Type hint sanified (i.e., sanitized) from a possibly insane type hint
        into a hopefully sane type hint by a
        :mod:`beartype._check.convert.convsanify` function.
    typevar_to_hint : TypeVarToHint
        **Type variable lookup table** describing this type hint, defined as an
        immutable dictionary mapping from the :pep:`484`-compliant **type
        variables** (i.e., :class:`typing.TypeVar` objects) originally
        parametrizing the origin of this type hint if any to the corresponding
        child type hints subscripting this type hint.
    _hash : int
        Hash identifying this object, precomputed for efficiency.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'hint',
        'typevar_to_hint',
        '_hash',
    )


    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        hint: Hint
        typevar_to_hint: TypeVarToHint


    _INIT_PARAM_NAMES = frozenset((
        var_name
        for var_name in __slots__
        # Ignore private slotted instance variables defined above.
        if not var_name.startswith('_')
    ))
    '''
    Frozen set of the names of all parameters accepted by the :meth:`init`
    method, defined as the frozen set comprehension of all public slotted
    instance variables of this class.

    This frozen set enables efficient membership testing.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Mandatory parameters.
        hint: Hint,

        # Optional parameters.
        typevar_to_hint: TypeVarToHint = FROZENDICT_EMPTY,
    ) -> None:
        '''
        Initialize this sanified type hint metadata with the passed parameters.

        Parameters
        ----------
        hint : Hint
            Type hint sanified (i.e., sanitized) from a possibly insane type
            hint into a hopefully sane type hint by a
            :mod:`beartype._check.convert.convsanify` function.
        typevar_to_hint : TypeVarToHint, optional
            **Type variable lookup table** describing this type hint, defined as
            an immutable dictionary mapping from the :pep:`484`-compliant **type
            variables** (i.e., :class:`typing.TypeVar` objects) originally
            parametrizing the origin of this type hint if any to the
            corresponding child type hints subscripting this type hint. Defaults
            to the empty table.
        '''
        assert isinstance(typevar_to_hint, FrozenDict), (
            f'{repr(typevar_to_hint)} not frozen dictionary.')

        # Classify all passed parameters as instance variables.
        self.hint = hint
        self.typevar_to_hint = typevar_to_hint

        # Hash identifying this object, precomputed for efficiency.
        self._hash = hash((hint, typevar_to_hint))

    # ..................{ DUNDERS                            }..................
    def __hash__(self) -> int:
        '''
        Hash identifying this sanified type hint metadata.

        Returns
        -------
        int
            This hash.
        '''

        return self._hash


    def __eq__(self, other: object) -> bool:
        '''
        :data:`True` only if this sanified type hint metadata is equal to the
        passed arbitrary object.

        Parameters
        ----------
        other : object
            Arbitrary object to be compared for equality against this metadata.

        Returns
        -------
        Union[bool, type(NotImplemented)]
            Either:

            * If this other object is also sanified type hint metadata, either:

              * If these metadatum share equal instance variables, :data:`True`.
              * Else, :data:`False`.

            * Else, :data:`NotImplemented`.
        '''

        # Return either...
        return (
            # If this other object is also sanified hint metadata, true only
            # if these metadatum share the same instance variables;
            (
                self.hint == other.hint and
                self.typevar_to_hint == other.typevar_to_hint
            )
            if isinstance(other, HintSanifiedData) else
            # Else, this other object is *NOT* also sanified hint metadata. In
            # this case, the standard singleton informing Python that this
            # equality comparator fails to support this comparison.
            NotImplemented  # type: ignore[return-value]
        )


    def __repr__(self) -> str:
        '''
        Machine-readable representation of this metadata.
        '''

        # Represent this metadata with just the minimal subset of metadata
        # needed to reasonably describe this metadata.
        return (
            f'{self.__class__.__name__}('
            f'hint={repr(self.hint)}, '
            f'typevar_to_hint={repr(self.typevar_to_hint)}'
            f')'
        )

    # ..................{ PERMUTERS                          }..................
    def permute(self, **kwargs) -> 'HintSanifiedData':
        '''
        Shallow copy of this metadata such that each passed keyword parameter
        overwrites the instance variable of the same name in this copy.

        Parameters
        ----------
        Keyword parameters of the same name and type as instance variables of
        this object (e.g., ``hint: Hint``, ``typevar_to_hint: TypeVarToHint``).

        Returns
        -------
        HintSanifiedData
            Shallow copy of this metadata such that each keyword parameter
            overwrites the instance variable of the same name in this copy.

        Raises
        ------
        _BeartypeDecorHintSanifyException
            If the name of any passed keyword parameter is *not* that of an
            existing instance variable of this object.
        '''

        # Set us up the permutation! Make your time!
        return permute_object(
            obj=self,
            init_arg_name_to_value=kwargs,
            init_arg_names=self._INIT_PARAM_NAMES,
            exception_cls=_BeartypeDecorHintSanifyException,
        )

# ....................{ HINTS                              }....................
HintOrSanifiedData = Union[Hint, HintSanifiedData]
'''
PEP-compliant type hint matching either a type hint *or* **sanified type hint
metadata** (i.e., :class:`.HintSanifiedData` object).
'''


HintOrSanifiedDataUnpacked = Tuple[Hint, TypeVarToHint]
'''
PEP-compliant type hint matching a 2-tuple ``(hint, typevar_to_hint)`` as
returned by the :func:`.unpack_hint_or_sane` unpacker.
'''

# ....................{ HINTS ~ container                  }....................
DictHintOrSanifiedDataToAny = Dict[HintOrSanifiedData, Any]
'''
PEP-compliant type hint matching a dictionary mapping from keys that are either
type hints *or* **sanified type hint metadata** (i.e.,
:class:`.HintSanifiedData` objects) to arbitrary objects.
'''

IterableHintOrSanifiedData = Iterable[HintOrSanifiedData]
'''
PEP-compliant type hint matching an iterable of zero or more items, each of
which is either a type hint *or* **sanified type hint metadata** (i.e.,
:class:`.HintSanifiedData` object).
'''


ListHintOrSanifiedData = List[HintOrSanifiedData]
'''
PEP-compliant type hint matching a list of zero or more items, each of which is
either a type hint *or* **sanified type hint metadata** (i.e.,
:class:`.HintSanifiedData` object).
'''


SetHintOrSanifiedData = Set[HintOrSanifiedData]
'''
PEP-compliant type hint matching a set of zero or more items, each of which is
either a type hint *or* **sanified type hint metadata** (i.e.,
:class:`.HintSanifiedData` object).
'''


TupleHintOrSanifiedData = Tuple[HintOrSanifiedData, ...]
'''
PEP-compliant type hint matching a tuple of zero or more items, each of which is
either a type hint *or* **sanified type hint metadata** (i.e.,
:class:`.HintSanifiedData` object).
'''

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_hint_or_sane_hint(hint_or_sane: HintOrSanifiedData) -> Hint:
    '''
    Hint unpacked (i.e., coerced, converted) from the passed parameter.

    Caveats
    -------
    This getter effectively discards *all* supplementary metadata recorded with
    this parameter -- including any type variable lookup table associated with
    this child hint.

    Parameters
    ----------
    hint_or_sane : HintSanifiedData
        Either a type hint *or* **sanified type hint metadata** (i.e.,
        :class:`.HintSanifiedData` object) to be unpacked.

    Returns
    -------
    Hint
        Hint unpacked (i.e., coerced, converted) from this parameter.
    '''

    # Return the hint encapsulated by this metadata.
    return (
        hint_or_sane.hint
        if isinstance(hint_or_sane, HintSanifiedData) else
        hint_or_sane
    )

# ....................{ UNPACKERS                          }....................
#FIXME: Unit test us up, please.
def unpack_hint_or_sane(
    # Mandatory parameters.
    hint_or_sane: HintOrSanifiedData,

    # Optional parameters.
    typevar_to_hint: TypeVarToHint = FROZENDICT_EMPTY,
) -> HintOrSanifiedDataUnpacked:
    '''
    2-tuple ``(hint, typevar_to_hint)`` unpacked from the passed parameters.

    This low-level utility function "unpacks" (i.e., coerces, converts) the
    passed type hint and type variable lookup table.

    Parameters
    ----------
    hint_or_sane : HintSanifiedData
        Either a type hint *or* **sanified type hint metadata** (i.e.,
        :class:`.HintSanifiedData` object) to be unpacked.
    typevar_to_hint : TypeVarToHint
        **Type variable lookup table** (i.e., immutable dictionary mapping from
        the :pep:`484`-compliant **type variables** (i.e.,
        :class:`typing.TypeVar` objects) originally parametrizing the origins of
        all transitive parent hints of this hint to the corresponding child
        hints subscripting these parent hints). Defaults to
        :class:`.FROZENDICT_EMPTY`.

    Returns
    -------
    Tuple[Hint, TypeVarToHint]
        2-tuple ``(hint, typevar_to_hint)`` where:

        * ``hint`` is the hint encapsulated by the passed parameters.
        * ``typevar_to_hint`` is the type variable lookup table encapsulated by
          the passed parameters. If the pair of type variable lookup tables
          encapsulated by both the passed ``typevar_to_hint`` parameter *and*
          ``hint_or_sane.typevar_to_hint`` instance variable are non-empty, then
          this ``typevar_to_hint`` item of this 2-tuple is a new type variable
          lookup table efficiently merging this parameter and instance variable
          (in that order).
    '''
    assert isinstance(typevar_to_hint, FrozenDict), (
        f'{repr(typevar_to_hint)} not frozen dictionary.')

    # If reducing this hint generated supplementary metadata...
    if isinstance(hint_or_sane, HintSanifiedData):
        # This lower-level hint reduced from this higher-level hint.
        hint = hint_or_sane.hint

        # If reducing this hint generated a non-empty type variable lookup
        # table...
        if hint_or_sane.typevar_to_hint:
            # If the caller passed an empty type variable lookup table,
            # trivially replace the latter with the former.
            if not typevar_to_hint:
                typevar_to_hint = hint_or_sane.typevar_to_hint
            else:
                # Full type variable lookup table uniting...
                typevar_to_hint = (
                    # The type variable lookup table describing all
                    # transitive parent hints of this reduced hint *AND*...
                    typevar_to_hint |  # type: ignore[operator]
                    # The type variable lookup table describing this hint.
                    #
                    # Note that this table is intentionally the second
                    # rather than first operand of this "|" operation,
                    # efficiently ensuring that type variables mapped by
                    # this hint take precedence over type variables mapped
                    # by transitive parent hints of this hint.
                    hint_or_sane.typevar_to_hint
                )
        # Else, reducing this hint generated an empty type variable lookup
        # table. In this case, this table is ignorable.
    # Else, reducing this hint did *NOT* generate supplementary metadata. In
    # this case, preserve this reduced hint as is.
    else:
        hint = hint_or_sane

    # Return the 2-tuple containing this hint and type variable lookup table.
    return (hint, typevar_to_hint)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **type-checking code classes** (i.e., low-level classes storing
metadata describing each iteration of the breadth-first search (BFS) dynamically
generating pure-Python code snippets type-checking arbitrary objects against
PEP-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import TYPE_CHECKING
from beartype._check.code.snip.codesnipcls import HINT_INDEX_TO_HINT_PLACEHOLDER
from beartype._data.hint.datahintpep import (
    Hint,
    TypeVarToHint,
)
from beartype._data.hint.datahinttyping import FrozenSetInts
from beartype._util.kind.map.utilmapfrozen import FrozenDict
from beartype._util.utilobject import SENTINEL

# ....................{ DATACLASSES                        }....................
#FIXME: Unit test us up, please.
class HintMeta(object):
    '''
    **Type hint type-checking metadata** (i.e., low-level dataclass storing
    metadata describing the possibly nested type hint visited by the current
    iteration of the breadth-first search (BFS) dynamically generating
    pure-Python type-checking code snippets in the
    :func:`beartype._check.code.codemake.make_check_expr` factory).

    Attributes
    ----------
    hint : Hint
        Type hint currently visited by this BFS.
    hint_placeholder : str
        **Type-checking placeholder substring** to be globally replaced in the
        **type-checking wrapper function code snippet** (i.e., the
        ``func_wrapper_code`` local defined by the
        :func:`beartype._check.code.codemake.make_check_expr` factory) by a
        Python code snippet type-checking the **current pith expression** (i.e.,
        the ``pith_var_name`` local) against the **currently visited type hint**
        (i.e., the :attr:`hint` instance variable).
    indent_level : int
        **Indentation level** (i.e., 1-based positive integer providing the
        level of indentation appropriate for this hint).
        Indexing the
        :obj:`beartype._data.code.datacodeindent.INDENT_LEVEL_TO_CODE`
        dictionary singleton by this integer efficiently yields the current
        **indendation string** suitable for prefixing each line of code
        type-checking the current pith against this hint.
    recursable_hint_ids : FrozenSetInts
        **Recursion guard** (i.e., frozen set of the integers uniquely
        identifying all transitive recursable parent hints (i.e., supporting
        recursion) of this hint). If the integer identifying a subsequently
        visited child hint subscripting this hint already resides in this
        recursion guard, that child hint has already been visited by prior
        iteration and is thus a recursive hint. Since recursive hints are valid
        (rather than constituting an unexpected error), the caller is expected
        to detect this use case and silently short-circuit infinite recursion by
        avoiding revisiting that previously visited recursive child hint.
    pith_expr : str
        **Pith expression** (i.e., Python code snippet evaluating to the value
        of) the current **pith** (i.e., possibly nested object of the passed
        parameter or return to be type-checked against this hint). Note that
        this expression is intentionally *not* an assignment expression but
        rather the original inefficient expression provided by the parent type
        hint of this hint.
    pith_var_name_index : int
        **Pith variable name index** (i.e., 0-based integer suffixing the name
        of each local variable assigned the value of the current pith in an
        assignment expression, thus uniquifying this variable in the body of the
        current wrapper function). Indexing the
        :obj:`beartype._check.code.snip.codesnipcls.PITH_INDEX_TO_VAR_NAME`
        dictionary singleton by this integer efficiently yields the current
        **pith variable name** locally storing the value of the current pith.
    typevar_to_hint : TypeVarToHint
        **Type variable lookup table** (i.e., immutable dictionary mapping
        from the **type variables** (i.e., :pep:`484`-compliant
        :class:`typing.TypeVar` objects) originally parametrizing the
        origins of all transitive parent hints of this hint to the
        corresponding child hints subscripting those parent hints). This table
        enables :func:`beartype.beartype` to efficiently reduce a proper subset
        of type variables to non-type variables at decoration time, including:

        * :pep:`484`- or :pep:`585`-compliant **subscripted generics.** For
          example, this table enables runtime type-checkers to reduce the
          semantically useless pseudo-superclass ``list[T]`` to the
          semantically useful pseudo-superclass ``list[int]`` at decoration time
          in the following example:

          .. code-block:: python

             class MuhGeneric[T](list[T]): pass

             @beartype
             def muh_func(muh_arg: MuhGeneric[int]) -> None: pass

        * :pep:`695`-compliant **subscripted type aliases.** For example, this
          table enables runtime type-checkers to reduce the semantically useless
          type hint ``muh_type_alias[float]`` to the semantically useful type
          hint ``float | int`` at decoration time in the following example:

          .. code-block:: python

             type muh_type_alias[T] = T | int

             @beartype
             def muh_func(muh_arg: muh_type_alias[float]) -> None: pass
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'hint',
        'hint_placeholder',
        'indent_level',
        'pith_expr',
        'pith_var_name_index',
        'recursable_hint_ids',
        'typevar_to_hint',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        hint: Hint
        hint_placeholder: str
        indent_level: int
        pith_expr: str
        pith_var_name_index: int
        recursable_hint_ids: FrozenSetInts
        typevar_to_hint: TypeVarToHint

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, hint_index: int) -> None:
        '''
        Initialize this type-checking metadata.

        Parameters
        ----------
        hint_index : int
            0-based index of this type-checking metadata in the parent
            :class:`.HintsMeta` list containing this metadata.
        '''
        assert isinstance(hint_index, int), f'{repr(hint_index)} not integer.'
        assert hint_index >= 0, f'{repr(hint_index)} < 0.'

        # Placeholder string to be globally replaced by code type-checking the
        # current pith against this hint.
        self.hint_placeholder = HINT_INDEX_TO_HINT_PLACEHOLDER[hint_index]

        # Nullify all remaining instance variables for safety.
        self.hint = SENTINEL  # type: ignore[assignment]
        self.indent_level = SENTINEL  # type: ignore[assignment]
        self.pith_expr = SENTINEL  # type: ignore[assignment]
        self.pith_var_name_index = SENTINEL  # type: ignore[assignment]
        self.recursable_hint_ids = SENTINEL  # type: ignore[assignment]
        self.typevar_to_hint = SENTINEL  # type: ignore[assignment]


    def reinit(
        self,
        hint: Hint,
        indent_level: int,
        pith_expr: str,
        pith_var_name_index: int,
        recursable_hint_ids: FrozenSetInts,
        typevar_to_hint: TypeVarToHint,
    ) -> None:
        '''
        Reinitialize this type-checking metadata to reflect a new type hint.

        Parameters
        ----------
        hint : Hint
            Type hint to describe by this metadata.
        indent_level : int
            1-based indentation level describing the current level of
            indentation appropriate for this hint.
        recursable_hint_ids : FrozenSetInts
            **Recursion guard** (i.e., frozen set of the integers uniquely
            identifying all transitive recursable parent hints (i.e., supporting
            recursion) of this hint).
        pith_expr : str
            Python code snippet evaluating to the child pith to be type-checked
            against this hint.
        pith_var_name_index : int
            0-based integer suffixing the name of each local variable assigned
            the value of the current pith in an assignment expression.
        typevar_to_hint : TypeVarToHint
            **Type variable lookup table** (i.e., immutable dictionary mapping
            from the **type variables** (i.e., :pep:`484`-compliant
            :class:`typing.TypeVar` objects) originally parametrizing the
            origins of all transitive parent hints of this hint to the
            corresponding child hints subscripting those parent hints).

        See Also
        --------
        Class docstring for further details on the passed parameters.
        '''
        assert isinstance(indent_level, int), (
            f'{repr(indent_level)} not integer.')
        assert isinstance(pith_expr, str), (
            f'{repr(pith_expr)} not string.')
        assert isinstance(pith_var_name_index, int), (
            f'{repr(pith_var_name_index)} not integer.')
        assert isinstance(recursable_hint_ids, frozenset), (
            f'{repr(recursable_hint_ids)} not frozen set.')
        assert isinstance(typevar_to_hint, FrozenDict), (
            f'{repr(typevar_to_hint)} not frozen dictionary.')
        assert indent_level >= 1, f'{repr(indent_level)} < 1.'
        assert pith_expr, f'{repr(pith_expr)} empty.'
        assert pith_var_name_index >= 0, f'{repr(pith_var_name_index)} < 0.'

        # Classify all passed parameters.
        self.hint = hint
        self.indent_level = indent_level
        self.pith_expr = pith_expr
        self.pith_var_name_index = pith_var_name_index
        self.recursable_hint_ids = recursable_hint_ids
        self.typevar_to_hint = typevar_to_hint

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        '''
        Machine-readable representation of this metadata.
        '''

        # Represent this metadata with just the minimal subset of metadata
        # needed to reasonably describe this metadata.
        return (
            f'{self.__class__.__name__}('
            f'hint={repr(self.hint)}, '
            f'indent_level={repr(self.indent_level)}, '
            f'pith_expr={repr(self.pith_expr)}, '
            f'pith_var_name_index={repr(self.pith_var_name_index)}, '
            f'recursable_hint_ids={repr(self.recursable_hint_ids)}, '
            f'typevar_to_hint={repr(self.typevar_to_hint)}'
            f')'
        )
